#
#   Base network structure for Darknet networks
#   Copyright EAVISE
#

import os
import collections
import logging
import numpy as np
import torch
import torch.nn as nn
from ._lightnet import Lightnet
from ..layer._darknet import *

__all__ = ['Darknet']
log = logging.getLogger(__name__)


class Darknet(Lightnet):
    """ This network module provides functionality to load darknet weight files.
    """
    def __init__(self):
        super().__init__()
        self.header = [0, 2, 0]

    def load(self, weights_file, *args, **kwargs):
        """ This function will load the weights from a file.
        If the file extension is _.pt_, it will be considered as a `pytorch pickle file <http://pytorch.org/docs/stable/notes/serialization.html#recommended-approach-for-saving-a-model>`_.
        Otherwise, the file is considered to be a darknet binary weight file.

        Args:
            weights_file (str): path to file
            *args, \*\*kwargs: Extra arguments passed to :func:`lightnet.network.module.Lightnet.load` when loading pytorch weights

        Note:
            Darknet weight files also contain the number of images the network has been trained on. |br|
            In Lightnet however, this is a parameter from the loss function and as such this value cannot be correctly set on that object.
            This value will thus be ignored by lightnet and when saving a darknet file, this value will be set to zero.
        """
        if os.path.splitext(weights_file)[1] == '.pt':
            log.debug('Loading weights from pytorch file')
            super().load(weights_file, *args, **kwargs)
        else:
            log.debug('Loading weights from darknet file')
            self._load_darknet_weights(weights_file)

    def save(self, weights_file, *args, **kwargs):
        """ This function will save the weights to a file.
        If the file extension is ``.pt``, it will be considered as a `pytorch pickle file <http://pytorch.org/docs/stable/notes/serialization.html#recommended-approach-for-saving-a-model>`_.
        Otherwise, the file is considered to be a darknet binary weight file.

        Args:
            weights_file (str): path to file
            *args, \*\*kwargs: Extra arguments passed to :func:`lightnet.network.module.Lightnet.save` when saving as pytorch weights
        """
        if os.path.splitext(weights_file)[1] == '.pt':
            log.debug('Saving weights to pytorch file')
            super().save(weights_file, *args, **kwargs)
        else:
            log.debug('Saving weights to darknet file')
            self._save_darknet_weights(weights_file)

    def _load_darknet_weights(self, weights_file):
        weights = WeightLoader(weights_file)
        self.header = weights.header

        done_loading = False
        for name, module in self.named_layer_loop():
            if not done_loading:
                try:
                    weights.load_layer(module)
                    log.debug(f'Layer loaded: {name}')
                    if weights.start >= weights.size:
                        log.debug(f'Finished loading weights [{weights.start}/{weights.size} weights]')
                        done_loading = True
                except NotImplementedError:
                    log.debug(f'Layer skipped: {name} [{module.__class__.__name__}]')
            else:
                log.warning(f'No more weigths for layer: {name}')

        if not done_loading:
            log.debug(f'Finished loading weights [{weights.start}/{weights.size} weights]')

    def _save_darknet_weights(self, weights_file):
        weights = WeightSaver(self.header, 0)

        for name, module in self.named_layer_loop():
            try:
                weights.save_layer(module)
                log.debug(f'Layer saved: {name}')
            except NotImplementedError:
                log.debug(f'Layer skipped: {name} [{module.__class__.__name__}]')

        weights.write_file(weights_file)


class WeightLoader:
    """ Load darknet weight files into pytorch layers """
    def __init__(self, filename):
        with open(filename, 'rb') as fp:
            self.header = np.fromfile(fp, count=3, dtype=np.int32).tolist()
            ver_num = self.header[0]*100+self.header[1]*10+self.header[2]
            log.debug(f'Loading weight file: version {self.header[0]}.{self.header[1]}.{self.header[2]}')

            if ver_num <= 19:
                log.warning('Weight file uses sizeof to compute variable size, which might lead to undefined behaviour. (choosing int=int32, float=float32)')
                self.seen = int(np.fromfile(fp, count=1, dtype=np.int32)[0])
            elif ver_num <= 29:
                log.warning('Weight file uses sizeof to compute variable size, which might lead to undefined behaviour. (choosing int=int32, float=float32, size_t=int64)')
                self.seen = int(np.fromfile(fp, count=1, dtype=np.int64)[0])
            else:
                log.error('New weight file syntax! Loading of weights might not work properly. Please submit an issue with the weight file version number. [Run with DEBUG logging level]')
                self.seen = int(np.fromfile(fp, count=1, dtype=np.int64)[0])

            self.buf = np.fromfile(fp, dtype=np.float32)

        self.start = 0
        self.size = self.buf.size

    def load_layer(self, layer):
        """ Load weights for a layer from the weights file """
        if type(layer) == nn.Conv2d:
            self._load_conv(layer)
        elif type(layer) == Conv2dBatchReLU:
            self._load_convbatch(layer)
        elif type(layer) == nn.Linear:
            self._load_fc(layer)
        else:
            raise NotImplementedError(f'The layer you are trying to load is not supported [{type(layer)}]')

    def _load_conv(self, model):
        num_b = model.bias.numel()
        model.bias.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_b])
                                   .view_as(model.bias.data))
        self.start += num_b

        num_w = model.weight.numel()
        model.weight.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_w])
                                     .view_as(model.weight.data))
        self.start += num_w

    def _load_convbatch(self, model):
        num_b = model.layers[1].bias.numel()
        model.layers[1].bias.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_b])
                                             .view_as(model.layers[1].bias.data))
        self.start += num_b
        model.layers[1].weight.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_b])
                                               .view_as(model.layers[1].weight.data))
        self.start += num_b
        model.layers[1].running_mean.copy_(torch.from_numpy(self.buf[self.start:self.start+num_b])
                                                .view_as(model.layers[1].running_mean))
        self.start += num_b
        model.layers[1].running_var.copy_(torch.from_numpy(self.buf[self.start:self.start+num_b])
                                               .view_as(model.layers[1].running_var))
        self.start += num_b

        num_w = model.layers[0].weight.numel()
        model.layers[0].weight.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_w])
                                               .view_as(model.layers[0].weight.data))
        self.start += num_w

    def _load_fc(self, model):
        num_b = model.bias.numel()
        model.bias.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_b])
                                   .view_as(model.bias.data))
        self.start += num_b

        num_w = model.weight.numel()
        model.weight.data.copy_(torch.from_numpy(self.buf[self.start:self.start+num_w])
                                     .view_as(model.weight.data))
        self.start += num_w


class WeightSaver:
    """ Save darknet weight files from pytorch layers """
    def __init__(self, header, seen):
        self.weights = []
        self.header = np.array(header, dtype=np.int32)
        ver_num = self.header[0]*100+self.header[1]*10+self.header[2]
        if ver_num <= 19:
            self.seen = np.int32(seen)
        elif ver_num <= 29:
            self.seen = np.int64(seen)
        else:
            log.error('New weight file syntax! Saving of weights might not work properly. Please submit an issue with the weight file version number. [Run with DEBUG logging level]')
            self.seen = np.int64(seen)

    def write_file(self, filename):
        """ Save the accumulated weights to a darknet weightfile """
        log.debug(f'Writing weight file: version {self.header[0]}.{self.header[1]}.{self.header[2]}')
        with open(filename, 'wb') as fp:
            self.header.tofile(fp)
            self.seen.tofile(fp)
            for np_arr in self.weights:
                np_arr.tofile(fp)
        log.info(f'Weight file saved as {filename}')

    def save_layer(self, layer):
        """ save weights for a layer """
        if type(layer) == nn.Conv2d:
            self._save_conv(layer)
        elif type(layer) == Conv2dBatchReLU:
            self._save_convbatch(layer)
        elif type(layer) == nn.Linear:
            self._save_fc(layer)
        else:
            raise NotImplementedError(f'The layer you are trying to save is not supported [{type(layer)}]')

    def _save_conv(self, model):
        self.weights.append(model.bias.cpu().data.numpy())
        self.weights.append(model.weight.cpu().data.numpy())

    def _save_convbatch(self, model):
        self.weights.append(model.layers[1].bias.cpu().data.numpy())
        self.weights.append(model.layers[1].weight.cpu().data.numpy())
        self.weights.append(model.layers[1].running_mean.cpu().numpy())
        self.weights.append(model.layers[1].running_var.cpu().numpy())
        self.weights.append(model.layers[0].weight.cpu().data.numpy())

    def _save_fc(self, model):
        self.weights.append(model.bias.cpu().data.numpy())
        self.weights.append(model.weight.cpu().data.numpy())



#
#   Utility layers
#   Copyright EAVISE
#

from collections import OrderedDict
import logging
import torch
import torch.nn as nn


__all__ = ['Residual', 'SequentialSelect', 'Parallel', 'ParallelCat', 'ParallelSum']
log = logging.getLogger(__name__)


class Residual(nn.Sequential):
    """ Residual block that runs like a Sequential, but then adds the original input to the output tensor.
    See :class:`torch.nn.Sequential` for more information.

    Args:
        *args: Arguments passed to :class:`torch.nn.Sequential`
        skip (nn.Module, optional): Extra module that is run on the input before adding it to the main block; Default **None**
        post (nn.Module, optional): Extra module that is run on the output after everything is added; Default **None**

    .. figure:: /.static/api/residual.*
       :width: 100%
       :alt: Residual module design

    Note:
        If you are using an OrderedDict to pass the modules to the sequential,
        you can set the `skip` and `post` values inside of that dict as well.

    Example:
        >>> # Note that the input channels and output channels should be the same for each branch (ic. 3 and 32)
        >>> module = ln.network.layer.Residual(
        ...     ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
        ...     ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
        ...     ln.network.layer.Conv2dBatchReLU(64, 32, 3, 1, 1),
        ...     # The skip block should ensure that the output tensor has the same number of channels
        ...     skip=ln.network.layer.Conv2dBatchReLU(3, 32, 1, 1, 0),
        ...     # The post block should run on the summed tensor,
        ...     # so the in_channels are equal to the out_channels of the output of the residual
        ...     post=ln.network.layer.Conv2dBatchReLU(32, 1, 1, 1, 0)
        ... )
        >>> print(module)
        Residual(
          (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          (skip): Conv2dBatchReLU(3, 32, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), ReLU(inplace=True))
          (post): Conv2dBatchReLU(32, 1, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), ReLU(inplace=True))
        )
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 1, 10, 10])
    """
    def __init__(self, *args, skip=None, post=None):
        super().__init__(*args)
        if skip is None and 'skip' in dir(self):
            self.skip = self._modules['skip']
        else:
            self.skip = skip

        if post is None and 'post' in dir(self):
            self.post = self._modules['post']
        else:
            self.post = post

    def forward(self, x):
        y = x
        for name, module in self.named_children():
            if name not in ('skip', 'post'):
                y = module(y)

        if self.skip is not None:
            x = self.skip(x)

        z = x + y
        if self.post is not None:
            z = self.post(z)

        return z


class SequentialSelect(nn.Sequential):
    """ Sequential that allows to select which layers are to be considered as output.
        See :class:`torch.nn.Sequential` for more information.

        Args:
            selection (list): names of the layers for which you want to get the output
            return_selection (bool): Whether to return the selected layers, or just store them as `self.selected`
            *args: Arguments that are passed to the Sequential init function

        .. figure:: /.static/api/sequentialselect.*
           :width: 100%
           :alt: SequentialSelect module design

        Note:
            If your selection only contains one item, this layer will flatten the return output tensor:

            >>> main_output, selected_output = layer(input)  # doctest: +SKIP

            However, if you have multiple outputs in your selection list, the outputs will be grouped in a dictionary:

            >>> main_output, selected_output_dict = layer(input)  # doctest: +SKIP

            The same applies for `self.selected` in case `return_selection` is set to False.

        Example:
            >>> module = ln.network.layer.SequentialSelect(
            ...     # We want to return the output from layers '1' and '3'
            ...     [1, 3],
            ...
            ...     # Since we specify False, the selected outputs will not be returned,
            ...     # but we can access them as `module.selected`
            ...     False,
            ...
            ...     # Sequential
            ...     ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(64, 64, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(64, 32, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(64, 128, 3, 1, 1),
            ... )
            >>> print(module)
            SequentialSelect(
              selection=['1', '3'], return=False, flatten=False
              (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
              (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
              (2): Conv2dBatchReLU(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
              (3): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
              (4): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
              (5): Conv2dBatchReLU(64, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            )
            >>> in_tensor = torch.rand(1, 3, 10, 10)
            >>> out_tensor = module(in_tensor)
            >>> print(out_tensor.shape)
            torch.Size([1, 128, 10, 10])
            >>> print(module.selected['1'].shape)
            torch.Size([1, 64, 10, 10])
            >>> print(module.selected['3'].shape)
            torch.Size([1, 32, 10, 10])

            >>> # Setting return_selection to True means the module will return a tuple of (output, selection)
            >>> module = ln.network.layer.SequentialSelect(
            ...     [1, 3], True,
            ...     ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(64, 64, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(64, 32, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
            ...     ln.network.layer.Conv2dBatchReLU(64, 128, 3, 1, 1),
            ... )
            >>> in_tensor = torch.rand(1, 3, 10, 10)
            >>> out_tensor, selected = module(in_tensor)
            >>> print(out_tensor.shape)
            torch.Size([1, 128, 10, 10])
            >>> print(selected['1'].shape)
            torch.Size([1, 64, 10, 10])
            >>> print(selected['3'].shape)
            torch.Size([1, 32, 10, 10])
    """
    def __init__(self, selection, return_selection, *args):
        super().__init__(*args)

        self.return_selection = return_selection
        self.selected = None
        self.selection = [str(select) for select in selection]
        self.flatten = len(self.selection) == 1
        k = list(self._modules.keys())
        for sel in self.selection:
            if sel not in k:
                raise KeyError(f'Selection key not found in sequential [{sel}]')

    def extra_repr(self):
        return f'selection={self.selection}, return={self.return_selection}, flatten={self.flatten}'

    def forward(self, x):
        sel_output = {sel: None for sel in self.selection}

        for key, module in self._modules.items():
            x = module(x)
            if key in self.selection:
                sel_output[key] = x

        # Return
        if not self.return_selection:
            if self.flatten:
                self.selected = sel_output[self.selection[0]]
            else:
                self.selected = sel_output

            return x

        if self.flatten:
            return x, sel_output[self.selection[0]]
        else:
            return x, sel_output


class Parallel(nn.Sequential):
    """ Container that runs each module on the input.
    The ouput is a list that contains the output of each of the different modules.

    Args:
        *args: Modules to run in parallel (similar to :class:`torch.nn.Sequential`)

    .. figure:: /.static/api/parallel.*
       :width: 100%
       :alt: Parallel module design

    Example:
        >>> # Note that the input channels should be the same for each branch (ic. 3)
        >>> module = ln.network.layer.Parallel(
        ...     ln.network.layer.Conv2dBatchReLU(3, 16, 3, 1, 1),
        ...     torch.nn.Sequential(
        ...         ln.network.layer.Conv2dBatchReLU(3, 8, 3, 1, 1),
        ...         ln.network.layer.Conv2dBatchReLU(8, 16, 3, 1, 1),
        ...         ln.network.layer.Conv2dBatchReLU(16, 32, 3, 1, 1),
        ...     ),
        ...     ln.network.layer.InvertedBottleneck(3, 64, 3, 1, 1),
        ... )
        >>> print(module)
        Parallel(
          (0): Conv2dBatchReLU(3, 16, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          (1): Sequential(
            (0): Conv2dBatchReLU(3, 8, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(8, 16, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(16, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
          (2): InvertedBottleneck(3, 64, kernel_size=(3, 3), stride=(1, 1), expansion=1, ReLU(inplace=True))
        )
        >>>
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out1, out2, out3 = module(in_tensor)
        >>> out1.shape
        torch.Size([1, 16, 10, 10])
        >>> out2.shape
        torch.Size([1, 32, 10, 10])
        >>> out3.shape
        torch.Size([1, 64, 10, 10])
    """
    def __init__(self, *args):
        super().__init__(*args)

    def forward(self, x):
        return [module(x) for module in self]


class ParallelCat(nn.Sequential):
    """ Parallel container that runs each module on the input and combines the different outputs by concatenating them.
    The tensors are considered as batched tensors and are thus concatenated in dimension 1.

    Args:
        *args: Arguments passed to :class:`~lightnet.network.layer._util.Parallel`
        post (nn.Module, optional): Extra module that is run on the sum of the outputs of the other modules; Default **None**

    .. figure:: /.static/api/parallelcat.*
       :width: 100%
       :alt: ParallelCat module design

    Note:
        If you are using an `OrderedDict` to pass the modules to the sequential,
        you can set the `post` value inside of that dict as well.

    Example:
        >>> # Note that the input channels should be the same for each branch (ic. 3)
        >>> module = ln.network.layer.ParallelCat(
        ...     ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
        ...     torch.nn.Sequential(
        ...         ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
        ...         ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
        ...         ln.network.layer.Conv2dBatchReLU(64, 32, 3, 1, 1),
        ...     ),
        ...     ln.network.layer.InvertedBottleneck(3, 32, 3, 1, 1),
        ...
        ...     # The post block should run on the concatenated tensor,
        ...     # so the in_channels are equal to the sum of the out_channels of the parallel modules
        ...     post=ln.network.layer.Conv2dBatchReLU(96, 32, 1, 1, 0)
        ... )
        >>> print(module)
        ParallelCat(
          (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          (1): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
          (2): InvertedBottleneck(3, 32, kernel_size=(3, 3), stride=(1, 1), expansion=1, ReLU(inplace=True))
          (post): Conv2dBatchReLU(96, 32, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), ReLU(inplace=True))
        )
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 10, 10])
    """
    def __init__(self, *args, post=None):
        super().__init__(*args)

        if post is None and 'post' in dir(self):
            self.post = self._modules['post']
        else:
            self.post = post

    def forward(self, x):
        output = torch.cat([module(x) for name, module in self.named_children() if name != 'post'], dim=1)
        if self.post is not None:
            output = self.post(output)

        return output


class ParallelSum(nn.Sequential):
    """ Parallel container that runs each module on the input and combines the different outputs by summing them.

    Args:
        *args: Arguments passed to :class:`~lightnet.network.layer._util.Parallel`
        post (nn.Module, optional): Extra module that is run on the sum of the outputs of the other modules; Default **None**

    .. figure:: /.static/api/parallelsum.*
       :width: 100%
       :alt: ParallelSum module design

    Note:
        If you are using an `OrderedDict` to pass the modules to the sequential,
        you can set the `post` value inside of that dict as well.

    Example:
        >>> # Note that the input channels and output channels should be the same for each branch (ic. 3 and 32)
        >>> module = ln.network.layer.ParallelSum(
        ...     ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
        ...     torch.nn.Sequential(
        ...         ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1),
        ...         ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
        ...         ln.network.layer.Conv2dBatchReLU(64, 32, 3, 1, 1),
        ...     ),
        ...     ln.network.layer.InvertedBottleneck(3, 32, 3, 1, 1),
        ...
        ...     # The post block should run on the summed tensor,
        ...     # so the in_channels are equal to the out_channels of the parallel modules
        ...     post=ln.network.layer.Conv2dBatchReLU(32, 1, 1, 1, 0),
        ... )
        >>> print(module)
        ParallelSum(
          (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          (1): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
          (2): InvertedBottleneck(3, 32, kernel_size=(3, 3), stride=(1, 1), expansion=1, ReLU(inplace=True))
          (post): Conv2dBatchReLU(32, 1, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), ReLU(inplace=True))
        )
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 1, 10, 10])
    """
    def __init__(self, *args, post=None):
        super().__init__(*args)

        if post is None and 'post' in dir(self):
            self.post = self._modules['post']
        else:
            self.post = post

    def forward(self, x):
        output = torch.sum(torch.stack([module(x) for name, module in self.named_children() if name != 'post']), dim=0)
        if self.post is not None:
            output = self.post(output)

        return output



#
#   Lightnet data augmentation
#   Copyright EAVISE
#

import random
import math
import numpy as np
import torch
from ..util import BaseTransform, BaseMultiTransform
from ..._imports import cv2, Image

__all__ = ['RandomFlip', 'RandomHSV', 'RandomJitter', 'RandomRotate']


class RandomFlip(BaseMultiTransform):
    """ Randomly flip image.

    Args:
        horizontal (Number [0-1]): Chance of flipping the image horizontally
        vertical (Number [0-1], optional): Chance of flipping the image vertically; Default **0**

    Note:
        Create 1 RandomFlip object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.
    """
    def __init__(self, horizontal, vertical=0):
        super().__init__()
        self.horizontal = horizontal
        self.vertical = vertical
        self.flip_h = False
        self.flip_v = False
        self.im_w = None
        self.im_h = None

    def _get_params(self):
        self.flip_h = random.random() < self.horizontal
        self.flip_v = random.random() < self.vertical

    def _tf_pil(self, img):
        self._get_params()
        self.im_w, self.im_h = img.size

        if self.flip_h and self.flip_v:
            img = img.transpose(Image.ROTATE_180)
        elif self.flip_h:
            img = img.transpose(Image.FLIP_LEFT_RIGHT)
        elif self.flip_v:
            img = img.transpose(Image.FLIP_TOP_BOTTOM)

        return img

    def _tf_cv(self, img):
        self._get_params()
        self.im_h, self.im_w = img.shape[:2]

        if self.flip_h and self.flip_v:
            img = cv2.flip(img, -1)
        elif self.flip_h:
            img = cv2.flip(img, 1)
        elif self.flip_v:
            img = cv2.flip(img, 0)

        return img

    def _tf_torch(self, img):
        self._get_params()

        if self.flip_h and self.flip_v:
            img = torch.flip(img, (1, 2))
        elif self.flip_h:
            img = torch.flip(img, (2,))
        elif self.flip_v:
            img = torch.flip(img, (1,))

        return img

    def _tf_anno(self, anno):
        anno = anno.copy()

        if self.flip_h and self.im_w is not None:
            anno.x_top_left = self.im_w - anno.x_top_left - anno.width
        if self.flip_v and self.im_h is not None:
            anno.y_top_left = self.im_h - anno.y_top_left - anno.height

        return anno


class RandomHSV(BaseTransform):
    """ Perform random HSV shift on the RGB data.

    Args:
        hue (Number): Random number between -hue,hue is used to shift the hue
        saturation (Number): Random number between 1,saturation is used to shift the saturation; 50% chance to get 1/dSaturation in stead of dSaturation
        value (Number): Random number between 1,value is used to shift the value; 50% chance to get 1/dValue in stead of dValue

    Warning:
        If you use OpenCV as your image processing library, make sure the image is RGB before using this transform.
        By default OpenCV uses BGR, so you must use `cvtColor`_ function to transform it to RGB.

    .. _cvtColor: https://docs.opencv.org/3.4/d8/d01/group__imgproc__color__conversions.html#ga397ae87e1288a81d2363b61574eb8cab
    """
    def __init__(self, hue, saturation, value):
        super().__init__()
        self.hue = hue
        self.saturation = saturation
        self.value = value

    def _get_params(self):
        self.dh = random.uniform(-self.hue, self.hue)

        self.ds = random.uniform(1, self.saturation)
        if random.random() < 0.5:
            self.ds = 1 / self.ds

        self.dv = random.uniform(1, self.value)
        if random.random() < 0.5:
            self.dv = 1 / self.dv

    def _tf_pil(self, img):
        self._get_params()
        img = img.convert('HSV')
        channels = list(img.split())

        def wrap_hue(x):
            x += int(self.dh * 255)
            if x > 255:
                x -= 255
            elif x < 0:
                x += 255
            return x

        channels[0] = channels[0].point(wrap_hue)
        channels[1] = channels[1].point(lambda i: min(255, max(0, int(i*self.ds))))
        channels[2] = channels[2].point(lambda i: min(255, max(0, int(i*self.dv))))

        img = Image.merge(img.mode, tuple(channels))
        img = img.convert('RGB')
        return img

    def _tf_cv(self, img):
        self._get_params()
        img = img.astype(np.float32) / 255.0
        img = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)

        img[:, :, 0] = self.wrap_hue(img[:, :, 0] + (360.0 * self.dh))
        img[:, :, 1] = np.clip(self.ds * img[:, :, 1], 0.0, 1.0)
        img[:, :, 2] = np.clip(self.dv * img[:, :, 2], 0.0, 1.0)

        img = cv2.cvtColor(img, cv2.COLOR_HSV2RGB)
        img = (img * 255).astype(np.uint8)
        return img

    def _tf_torch(self, img):
        self._get_params()

        # Transform to HSV
        maxval, _ = img.max(0)
        minval, _ = img.min(0)
        diff = maxval - minval

        h = torch.zeros_like(diff)
        mask = (diff != 0) & (maxval == img[0])
        h[mask] = (60 * (img[1, mask] - img[2, mask]) / diff[mask] + 360)
        mask = (diff != 0) & (maxval == img[1])
        h[mask] = (60 * (img[2, mask] - img[0, mask]) / diff[mask] + 120)
        mask = (diff != 0) & (maxval == img[2])
        h[mask] = (60 * (img[0, mask] - img[1, mask]) / diff[mask] + 240)
        h %= 360

        s = torch.zeros_like(diff)
        mask = maxval != 0
        s[mask] = diff[mask] / maxval[mask]

        # Random Shift
        h = self.wrap_hue(h + (360 * self.dh))
        s = torch.clamp(self.ds * s, 0, 1)
        v = torch.clamp(self.dv * maxval, 0, 1)

        # Transform to RGB
        c = v * s
        m = v - c
        x = c * (1 - (((h / 60) % 2) - 1).abs())
        cm = c + m
        xm = x + m

        img = torch.stack((m, m, m))
        mask = (h >= 0) & (h <= 60)
        img[0, mask] = cm[mask]
        img[1, mask] = xm[mask]
        mask = (h > 60) & (h <= 120)
        img[0, mask] = xm[mask]
        img[1, mask] = cm[mask]
        mask = (h > 120) & (h <= 180)
        img[1, mask] = cm[mask]
        img[2, mask] = xm[mask]
        mask = (h > 180) & (h <= 240)
        img[1, mask] = xm[mask]
        img[2, mask] = cm[mask]
        mask = (h > 240) & (h <= 300)
        img[0, mask] = xm[mask]
        img[2, mask] = cm[mask]
        mask = (h > 300) & (h <= 360)
        img[0, mask] = cm[mask]
        img[2, mask] = xm[mask]

        return img

    @staticmethod
    def wrap_hue(h):
        h[h >= 360.0] -= 360.0
        h[h < 0.0] += 360.0
        return h


class RandomJitter(BaseMultiTransform):
    """ Add random jitter to an image, by randomly cropping (or adding borders) to each side.

    Args:
        jitter (Number [0-1]): Indicates how much of the image we can crop
        fill_color (int or float, optional): Fill color to be used for padding (if int, will be divided by 255); Default **0.5**

    Note:
        Create 1 RandomCrop object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.

    Warning:
        This transformation only modifies the annotations to fit the new origin point of the image.
        It does not crop the annotations to fit inside the new boundaries, nor does it filter annotations that fall outside of these new boundaries.
        Check out :class:`~lightnet.data.transform.FitAnno` for a transformation that does this.
    """
    def __init__(self, jitter, fill_color=0.5):
        super().__init__()
        self.jitter = jitter
        self.fill_color = fill_color if isinstance(fill_color, float) else fill_color / 255
        self.crop = None

    def _get_params(self, im_w, im_h):
        dw, dh = int(im_w*self.jitter), int(im_h*self.jitter)
        crop_left = random.randint(-dw, dw)
        crop_right = random.randint(-dw, dw)
        crop_top = random.randint(-dh, dh)
        crop_bottom = random.randint(-dh, dh)

        self.crop = (crop_left, crop_top, im_w-crop_right, im_h-crop_bottom)

    def _tf_pil(self, img):
        im_w, im_h = img.size
        self._get_params(im_w, im_h)
        crop_w = self.crop[2] - self.crop[0]
        crop_h = self.crop[3] - self.crop[1]
        shape = np.array(img).shape
        channels = shape[2] if len(shape) > 2 else 1

        img = img.crop((max(0, self.crop[0]), max(0, self.crop[1]), min(im_w, self.crop[2]), min(im_h, self.crop[3])))
        img_crop = Image.new(img.mode, (crop_w, crop_h), color=(int(self.fill_color*255),)*channels)
        img_crop.paste(img, (max(0, -self.crop[0]), max(0, -self.crop[1])))

        return img_crop

    def _tf_cv(self, img):
        im_h, im_w = img.shape[:2]
        self._get_params(im_w, im_h)

        crop_w = self.crop[2] - self.crop[0]
        crop_h = self.crop[3] - self.crop[1]
        img_crop = np.ones((crop_h, crop_w) + img.shape[2:], dtype=img.dtype) * int(self.fill_color*255)

        src_x1 = max(0, self.crop[0])
        src_x2 = min(self.crop[2], im_w)
        src_y1 = max(0, self.crop[1])
        src_y2 = min(self.crop[3], im_h)
        dst_x1 = max(0, -self.crop[0])
        dst_x2 = crop_w - max(0, self.crop[2]-im_w)
        dst_y1 = max(0, -self.crop[1])
        dst_y2 = crop_h - max(0, self.crop[3]-im_h)
        img_crop[dst_y1:dst_y2, dst_x1:dst_x2] = img[src_y1:src_y2, src_x1:src_x2]

        return img_crop

    def _tf_torch(self, img):
        im_h, im_w = img.shape[-2:]
        self._get_params(im_w, im_h)

        crop_w = self.crop[2] - self.crop[0]
        crop_h = self.crop[3] - self.crop[1]
        img_crop = torch.full((img.shape[0], crop_h, crop_w), self.fill_color, dtype=img.dtype)

        src_x1 = max(0, self.crop[0])
        src_x2 = min(self.crop[2], im_w)
        src_y1 = max(0, self.crop[1])
        src_y2 = min(self.crop[3], im_h)
        dst_x1 = max(0, -self.crop[0])
        dst_x2 = crop_w - max(0, self.crop[2]-im_w)
        dst_y1 = max(0, -self.crop[1])
        dst_y2 = crop_h - max(0, self.crop[3]-im_h)
        img_crop[:, dst_y1:dst_y2, dst_x1:dst_x2] = img[:, src_y1:src_y2, src_x1:src_x2]

        return img_crop

    def _tf_anno(self, anno):
        anno = anno.copy()

        anno.x_top_left -= self.crop[0]
        anno.y_top_left -= self.crop[1]

        return anno


class RandomRotate(BaseMultiTransform):
    """ Randomly rotate the image/annotations.
    For the annotations we take the smallest possible rectangle that fits the rotated rectangle.

    Args:
        jitter (Number [0-180]): Random number between -jitter,jitter degrees is used to rotate the image

    Note:
        Create 1 RandomRotate object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.
    """
    def __init__(self, jitter):
        super().__init__()
        self.jitter = jitter
        self.angle = None
        self.im_w = None
        self.im_h = None

    def _get_params(self, im_w, im_h):
        self.im_w = im_w
        self.im_h = im_h
        self.angle = random.randint(-self.jitter, self.jitter)

    def _tf_pil(self, img):
        im_w, im_h = img.size
        self._get_params(im_w, im_h)
        return img.rotate(self.angle)

    def _tf_cv(self, img):
        im_h, im_w = img.shape[:2]
        self._get_params(im_w, im_h)
        M = cv2.getRotationMatrix2D((im_w/2, im_h/2), self.angle, 1)
        return cv2.warpAffine(img, M, (im_w, im_h))

    def _tf_torch(self, img):
        raise NotImplementedError('Random Rotate is not implemented for torch Tensors, you can use Kornia [https://github.com/kornia/kornia]')

    def _tf_anno(self, anno):
        anno = anno.copy()

        cx, cy = self.im_w/2, self.im_h/2
        rad = math.radians(-self.angle)
        cos_a = math.cos(rad)
        sin_a = math.sin(rad)

        # Rotate anno
        x1_c = anno.x_top_left - cx
        y1_c = anno.y_top_left - cy
        x2_c = x1_c + anno.width
        y2_c = y1_c + anno.height

        x1_r = (x1_c * cos_a - y1_c * sin_a) + cx
        y1_r = (x1_c * sin_a + y1_c * cos_a) + cy
        x2_r = (x2_c * cos_a - y1_c * sin_a) + cx
        y2_r = (x2_c * sin_a + y1_c * cos_a) + cy
        x3_r = (x2_c * cos_a - y2_c * sin_a) + cx
        y3_r = (x2_c * sin_a + y2_c * cos_a) + cy
        x4_r = (x1_c * cos_a - y2_c * sin_a) + cx
        y4_r = (x1_c * sin_a + y2_c * cos_a) + cy
        rot_x = np.stack([x1_r, x2_r, x3_r, x4_r])
        rot_y = np.stack([y1_r, y2_r, y3_r, y4_r])

        # Max rect box
        anno.x_top_left = rot_x.min(axis=0)
        anno.y_top_left = rot_y.min(axis=0)
        anno.width = rot_x.max(axis=0) - anno.x_top_left
        anno.height = rot_y.max(axis=0) - anno.y_top_left

        return anno



#
#   Cornernet HourGlass Module
#   Copyright EAVISE
#

from collections import OrderedDict
import torch.nn as nn
from ._util import Residual, ParallelSum

__all__ = ['HourGlass']


class HourGlass(nn.Module):
    """ Fully modifiable HourGlass module. |br|
    The default implementation follows that of :cite:`hourglass`, with singular residual blocks between each group,
    which we adapted from `their repository <hgrepo_>`_ to the best of our abilities.
    However, you can completely overhaul this module to fit your needs,
    by supplying your own "make" functions for each of the seperate blocks.

    Args:
        order (int): The number of recursive modules
        channels (int or list of ints): Channels for each of the recursive modules
        make_upper (function, optional): Function that should return a :class:`torch.nn.Module` for the upper part; Default **ResidualBlock**
        upper_kwargs (dict or list of dicts): Extra kwargs that are passed to the `make_upper` function; Default **None**
        make_down1 (function, optional): Function that should return a :class:`torch.nn.Module` for the first downward part; Default **2x2 MaxPool + ResidualBlock**
        down1_kwargs (dict or list of dicts): Extra kwargs that are passed to the `make_down1` function; Default **None**
        make_inner (function, optional): Function that should return a :class:`torch.nn.Module` for the inner part; Default **ResidualBlock**
        inner_kwargs (dict or list of dicts): Extra kwargs that are passed to the `make_inner` function; Default **None**
        make_down2 (function, optional): Function that should return a :class:`torch.nn.Module` for the second downward part; Default **ResidualBlock + 2x2 Upsampling (NN)**
        down2_kwargs (dict or list of dicts): Extra kwargs that are passed to the `make_down2` function; Default **None**

    .. figure:: /.static/api/hourglass.*
       :width: 100%
       :alt: HourGlass module design

       The HourGlass module consists of 4 different blocks: upper, down1, inner, down2. |br|
       You can modify each of these blocks by providing the correct make function and (optionally) keyword arguments.

    Note:
        All "make" functions have the following signature: ``function(in_channels, out_channels, **kwargs)``.

        The `in_channels` and `out_channels` get computed depending on the block (up, down1, down2 or inner), the current `order` and corresponding number of `channels`. |br|
        The keyword arguments of the different "make" functions are taken from the matching `..._kwargs` arguments,
        which can either be a fixed dictionary which will be used each time the function gets called,
        or a list of dictionaries, with different values for each recursion of the HourGlass module.
        If you pass a list of dictionaries, it should contain `order` number of dictionaries.

    Note:
        If you pass a list of integers as `channels`, you need to enter `order + 1` integers.
        The last recursive HourGlass module, wil use this extra channel number for its `inner` module,
        allowing to use the same `down1` and `down2` modules in each HourGlass. |br|
        If you pass a single integer, it will be used for all modules.

    Example:
        >>> module = ln.network.layer.HourGlass(
        ...     2,                  # 2 nested hourglasses
        ...     [3, 32, 64],        # 2+1 channels, one for each hourglass level and a last one for the inner module
        ...
        ...     # UPPER: single convbatchrelu
        ...     make_upper=lambda ci, co: ln.network.layer.Conv2dBatchReLU(ci, co, 3, 1, 1),
        ...
        ...     # DOWN 1: Providing keyword arguments
        ...     make_down1=lambda ci, co, **kwargs: ln.network.layer.Conv2dBatchReLU(ci, co, **kwargs),
        ...     down1_kwargs={'kernel_size': 3, 'stride': 2, 'padding': 1},
        ...
        ...     # INNER: Conv2dDepthWise
        ...     make_inner=lambda ci, co: ln.network.layer.Conv2dDepthWise(ci, co, 3, 1, 1),
        ...
        ...     # DOWN 2: Separate kwargs for each hourglass level
        ...     make_down2=lambda ci, co, **kwargs: torch.nn.Sequential(
        ...         ln.network.layer.Conv2dBatchReLU(ci, co, **kwargs),
        ...         torch.nn.Upsample(scale_factor=2, mode='nearest'),
        ...     ),
        ...     down2_kwargs=[{'kernel_size': 3, 'stride': 1, 'padding': 1}, {'kernel_size': 1, 'stride': 1, 'padding': 0}],
        ... )
        >>> print(module)
        HourGlass(
          (layers): ParallelSum(
            (upper): Conv2dBatchReLU(3, 3, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (down): Sequential(
              (down1): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), ReLU(inplace=True))
              (inner): HourGlass(
                (layers): ParallelSum(
                  (upper): Conv2dBatchReLU(32, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
                  (down): Sequential(
                    (down1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), ReLU(inplace=True))
                    (inner): Conv2dDepthWise(64, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
                    (down2): Sequential(
                      (0): Conv2dBatchReLU(64, 32, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), ReLU(inplace=True))
                      (1): Upsample(scale_factor=2.0, mode=nearest)
                    )
                  )
                )
              )
              (down2): Sequential(
                (0): Conv2dBatchReLU(32, 3, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
                (1): Upsample(scale_factor=2.0, mode=nearest)
              )
            )
          )
        )

    .. _hgrepo: https://github.com/princeton-vl/pose-hg-train
    """
    def __init__(
        self,
        order,
        channels,
        make_upper=None,
        upper_kwargs=None,
        make_down1=None,
        down1_kwargs=None,
        make_inner=None,
        inner_kwargs=None,
        make_down2=None,
        down2_kwargs=None,
    ):
        super().__init__()

        make_upper = make_upper if make_upper is not None else self._make_upper
        make_down1 = make_down1 if make_down1 is not None else self._make_down1
        make_inner = make_inner if make_inner is not None else self._make_inner
        make_down2 = make_down2 if make_down2 is not None else self._make_down2
        c1 = self._get_current(channels)
        c2 = self._get_current(self._get_next(channels))

        if order > 1:
            inner = HourGlass(
                order - 1,
                self._get_next(channels),
                make_upper,
                self._get_next(upper_kwargs),
                make_down1,
                self._get_next(down1_kwargs),
                make_inner,
                self._get_next(inner_kwargs),
                make_down2,
                self._get_next(down2_kwargs),
            )
        else:
            inner = make_inner(c2, c2, **self._get_current(inner_kwargs, dict()))

        self.layers = ParallelSum(OrderedDict([
            ('upper', make_upper(c1, c1, **self._get_current(upper_kwargs, dict()))),
            ('down', nn.Sequential(OrderedDict([
                ('down1', make_down1(c1, c2, **self._get_current(down1_kwargs, dict()))),
                ('inner', inner),
                ('down2', make_down2(c2, c1, **self._get_current(down2_kwargs, dict()))),
            ]))),
        ]))

    def forward(self, x):
        return self.layers(x)

    @staticmethod
    def _make_upper(in_channels, out_channels, **kwargs):
        return get_residual(in_channels, out_channels)

    @staticmethod
    def _make_down1(in_channels, out_channels, **kwargs):
        return nn.Sequential(
            nn.MaxPool2d(2, 2),
            get_residual(in_channels, out_channels),
        )

    @staticmethod
    def _make_inner(in_channels, out_channels, **kwargs):
        return get_residual(in_channels, out_channels)

    @staticmethod
    def _make_down2(in_channels, out_channels, **kwargs):
        return nn.Sequential(
            get_residual(in_channels, out_channels),
            nn.Upsample(scale_factor=2, mode='nearest'),
        )

    @staticmethod
    def _get_current(value, default=None):
        if value is None:
            return default
        elif isinstance(value, (list, tuple)):
            return value[0] if len(value) >= 1 else default
        else:
            return value

    @staticmethod
    def _get_next(value):
        if isinstance(value, (list, tuple)):
            return value[1:]
        else:
            return value


def get_residual(in_channels, out_channels):
    """ Basic residual block implementation from :cite:`hourglass`. """
    return Residual(
        nn.BatchNorm2d(in_channels),
        nn.ReLU(inplace=True),
        nn.Conv2d(in_channels, out_channels//2, 1, 1, 0),
        nn.BatchNorm2d(out_channels//2),
        nn.ReLU(inplace=True),
        nn.Conv2d(out_channels//2, out_channels//2, 3, 1, 1),
        nn.BatchNorm2d(out_channels//2),
        nn.ReLU(inplace=True),
        nn.Conv2d(out_channels//2, out_channels, 1, 1, 0),

        skip=None if in_channels == out_channels else nn.Conv2d(in_channels, out_channels, 1, 1, 0)
    )



#
#   Dependency list related functions
#   Copyright EAVISE
#
import logging
import tempfile
import torch
from .._imports import onnx
from ._create import *
from ._tree import *
from ._type import *

__all__ = ['get_dependency_map', 'get_onnx_model', 'print_dependency_map']
log = logging.getLogger(__name__)


def get_dependency_map(model, input_dim):
    """ Computes the dependency map of a model.

    Args:
        model (torch.nn.Module): Network model
        input_dim (tuple <int>): Input dimension for the model, usually [batch, channels, height, width]

    Returns:
        dict: dependency map for each prunable convolution
    """
    # Get ONNX graph
    path = tempfile.TemporaryFile(prefix='lightnet-prune', suffix='.onnx')
    get_onnx_model(model, input_dim, path)
    path.seek(0)
    onnx_model = onnx.load_model(path, load_external_data=False)
    path.close()

    # Create dependency tree
    dependencies = dict()
    for el in onnx_model.graph.node:
        if el.op_type == 'ATen' and el.attribute[0].s == b'_convolution':
            try:
                name = [i for i in el.input if i.endswith('.weight')][0][:-7]
                module = model
                for p in name.split('.'):
                    module = getattr(module, p)
                if module.groups != 1:
                    raise NotImplementedError(f'ATen: Grouped Convolution (cannot prune)')

                dep = create_node(el, onnx_model.graph, model)
                dependencies[dep.name] = dep
            except NotImplementedError as err:
                log.info(f'Cannot prune [{name}], unimplemented dependency [{err}]')
            except StopIteration as err:
                log.info(f'Cannot prune [{name}], generates output [{err}]')

    # Remove ignored and match with modules
    for dep in dependencies.values():
        for _, node in traverse_depth_first(dep):
            if node.type is NodeType.IGNORE:
                for p in node.parents:
                    idx = p.children.index(node)
                    p.children[idx:idx+1] = node.children
                for c in node.children:
                    idx = c.parents.index(node)
                    c.parents[idx:idx+1] = node.parents
                del node
            elif node.name is not None:
                path = node.name.split('.')
                module = model
                try:
                    for p in path:
                        module = getattr(module, p)
                    node.module = module
                except AttributeError:
                    log.debug(f'Could not get PyTorch module for [{node.name}]')
                    continue

    return dependencies


def get_onnx_model(model, input_dim, path):
    """ Compute and save the ONNX version of a model. |br|
    This function computes the ONNX version of a model, which is used to compute the dependency map for pruning.

    Args:
        model (torch.nn.Module): Network model
        input_dim (tuple <int>): Input dimension for the model, usually [batch, channels, height, width]
        path (string): Path to save the ONNX model

    Note:
        This function is used internally by lightnet whilst computing the dependency map
        and should generally not be used by users.
        It is only made publicly available in order to help debug pruning issues.
    """
    # Get device
    try:
        device = next(model.parameters()).device
    except StopIteration:
        log.warn('Could not determine device from model, using "cpu".')
        device = torch.device('cpu')

    # Create input tensor
    input_tensor = torch.rand(*input_dim).to(device)

    # Create onnx model
    torch.onnx.export(
        model, input_tensor, path,
        keep_initializers_as_inputs=False,
        operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN,
        # Necessary because OperatorExportTypes.ONNX_ATEN
        do_constant_folding=False
    )


def print_dependency_map(dependencies, indentation=' '):
    """ Pretty print a dependency map for easier inspection.

    Args:
        dependencies (dict): The computed dependecy map (see :func:`~lightnet.prune.dependency.get_dependency_map`)
        indentation (string, optional): Indentation to add for each level of dependency; Default **' '**
    """
    for name, d in dependencies.items():
        print(name)
        for i, n in traverse_depth_first(d, True):
            print(f' {2 * i * indentation}{n}')
        print()



#
#   Lightnet : Darknet building blocks implemented in pytorch
#   Copyright EAVISE
#

__all__ = ['data', 'engine', 'models', 'network', 'prune']


from .version import __version__
from .log import *

from . import data
from . import engine
from . import models
from . import network
from . import prune



#
#   Lightnet preprocessing for fitting data to certain dimensions
#   Copyright EAVISE
#

import random
import collections
import numpy as np
import torch
from ..util import BaseMultiTransform
from ..._imports import cv2, Image, ImageOps

__all__ = ['Crop', 'Letterbox', 'Pad', 'FitAnno']


class Crop(BaseMultiTransform):
    """ Rescale and crop images/annotations to the right network dimensions.
    This transform will first rescale to the closest (bigger) dimension possible and then take a crop to the exact dimensions.

    Args:
        dimension (tuple, optional): Default size for the letterboxing, expressed as a (width, height) tuple; Default **None**
        dataset (lightnet.data.Dataset, optional): Dataset that uses this transform; Default **None**
        center (Boolean, optional): Whether to take the crop from the center or randomly.

    Note:
        Create 1 Crop object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.

    Warning:
        This transformation only modifies the annotations to fit the new scale and origin point of the image.
        It does not crop the annotations to fit inside the new boundaries, nor does it filter annotations that fall outside of these new boundaries.
        Check out :class:`~lightnet.data.transform.FitAnno` for a transformation that does this.
    """
    def __init__(self, dimension=None, dataset=None, center=True):
        super().__init__()
        self.dimension = dimension
        self.dataset = dataset
        self.center = center
        if self.dimension is None and self.dataset is None:
            raise ValueError('This transform either requires a dimension or a dataset to infer the dimension')

        self.scale = 1
        self.crop = None

    def _get_params(self, im_w, im_h):
        if self.dataset is not None:
            net_w, net_h = self.dataset.input_dim
        elif isinstance(self.dimension, int):
            net_w, net_h = self.dimension, self.dimension
        else:
            net_w, net_h = self.dimension

        if net_w / im_w >= net_h / im_h:
            self.scale = net_w / im_w
            xcrop = 0
            ycrop = int(im_h * self.scale - net_h + 0.5)
        else:
            self.scale = net_h / im_h
            xcrop = int(im_w * self.scale - net_w + 0.5)
            ycrop = 0

        if xcrop == 0 and ycrop == 0:
            self.crop = None
        else:
            dx = xcrop // 2 if self.center else random.randint(0, xcrop)
            dy = ycrop // 2 if self.center else random.randint(0, ycrop)
            self.crop = (dx, dy, dx + net_w, dy + net_h)

    def _tf_pil(self, img):
        im_w, im_h = img.size
        self._get_params(im_w, im_h)

        # Rescale
        if self.scale != 1:
            bands = img.split()
            bands = [b.resize((int(self.scale*im_w+0.5), int(self.scale*im_h+0.5)), resample=Image.BILINEAR) for b in bands]
            img = Image.merge(img.mode, bands)
            im_w, im_h = img.size

        # Crop
        if self.crop is not None:
            img = img.crop(self.crop)

        return img

    def _tf_cv(self, img):
        im_h, im_w = img.shape[:2]
        self._get_params(im_w, im_h)

        # Rescale
        if self.scale != 1:
            img = cv2.resize(img, (int(im_w*self.scale+0.5), int(im_h*self.scale+0.5)), interpolation=cv2.INTER_LINEAR)

        # Crop
        if self.crop is not None:
            img = img[self.crop[1]:self.crop[3], self.crop[0]:self.crop[2]]

        return img

    def _tf_torch(self, img):
        im_h, im_w = img.shape[-2:]
        self._get_params(im_w, im_h)

        # Rescale
        if self.scale != 1:
            if img.ndim == 3:
                img = img[None, ...]
            elif img.ndim == 2:
                img = img[None, None, ...]
            img = torch.nn.functional.interpolate(
                img,
                size=(int(im_h*self.scale+0.5), int(im_w*self.scale+0.5)),
                mode='bilinear',
                align_corners=False,
            ).squeeze().clamp(min=0, max=255)

        # Crop
        if self.crop is not None:
            img = img[..., self.crop[1]:self.crop[3], self.crop[0]:self.crop[2]]

        return img

    def _tf_anno(self, anno):
        anno = anno.copy()

        # Rescale
        if self.scale != 1:
            anno.x_top_left *= self.scale
            anno.y_top_left *= self.scale
            anno.width *= self.scale
            anno.height *= self.scale

        # Crop
        if self.crop is not None:
            anno.x_top_left -= self.crop[0]
            anno.y_top_left -= self.crop[1]

        return anno


class Letterbox(BaseMultiTransform):
    """ Rescale images/annotations and add top/bottom borders to get to the right network dimensions.

    Args:
        dimension (tuple, optional): Default size for the letterboxing, expressed as a (width, height) tuple; Default **None**
        dataset (lightnet.data.Dataset, optional): Dataset that uses this transform; Default **None**
        fill_color (int or float, optional): Fill color to be used for padding (if int, will be divided by 255); Default **0.5**

    Note:
        Create 1 Letterbox object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.
    """
    def __init__(self, dimension=None, dataset=None, fill_color=0.5):
        super().__init__()
        self.dimension = dimension
        self.dataset = dataset
        self.fill_color = fill_color if isinstance(fill_color, float) else fill_color / 255
        if self.dimension is None and self.dataset is None:
            raise ValueError('This transform either requires a dimension or a dataset to infer the dimension')

        self.pad = None
        self.scale = None

    def _get_params(self, im_w, im_h):
        if self.dataset is not None:
            net_w, net_h = self.dataset.input_dim
        elif isinstance(self.dimension, int):
            net_w, net_h = self.dimension, self.dimension
        else:
            net_w, net_h = self.dimension

        if im_w / net_w >= im_h / net_h:
            self.scale = net_w / im_w
        else:
            self.scale = net_h / im_h

        pad_w = (net_w - int(im_w*self.scale+0.5)) / 2
        pad_h = (net_h - int(im_h*self.scale+0.5)) / 2
        if pad_w == 0 and pad_h == 0:
            self.pad = None
        else:
            self.pad = (int(pad_w), int(pad_h), int(pad_w+.5), int(pad_h+.5))

    def _tf_pil(self, img):
        im_w, im_h = img.size
        self._get_params(im_w, im_h)

        # Rescale
        if self.scale != 1:
            bands = img.split()
            bands = [b.resize((int(self.scale*im_w+0.5), int(self.scale*im_h+0.5)), resample=Image.BILINEAR) for b in bands]
            img = Image.merge(img.mode, bands)

        # Pad
        if self.pad is not None:
            shape = np.array(img).shape
            channels = shape[2] if len(shape) > 2 else 1
            img = ImageOps.expand(img, border=self.pad, fill=(int(self.fill_color*255),)*channels)

        return img

    def _tf_cv(self, img):
        im_h, im_w = img.shape[:2]
        self._get_params(im_w, im_h)

        # Rescale
        if self.scale != 1:
            img = cv2.resize(img, (int(im_w*self.scale+0.5), int(im_h*self.scale+0.5)), interpolation=cv2.INTER_LINEAR)

        # Pad
        if self.pad is not None:
            channels = img.shape[2] if len(img.shape) > 2 else 1
            img = cv2.copyMakeBorder(img, self.pad[1], self.pad[3], self.pad[0], self.pad[2], cv2.BORDER_CONSTANT, value=(int(self.fill_color*255),)*channels)

        return img

    def _tf_torch(self, img):
        im_h, im_w = img.shape[-2:]
        self._get_params(im_w, im_h)

        # Rescale
        if self.scale != 1:
            if img.ndim == 3:
                img = img[None, ...]
            elif img.ndim == 2:
                img = img[None, None, ...]
            img = torch.nn.functional.interpolate(
                img,
                size=(int(im_h*self.scale+0.5), int(im_w*self.scale+0.5)),
                mode='bilinear',
                align_corners=False,
            ).squeeze().clamp(min=0, max=255)

        # Pad
        if self.pad is not None:
            img = torch.nn.functional.pad(img, (self.pad[0], self.pad[2], self.pad[1], self.pad[3]), value=self.fill_color)

        return img

    def _tf_anno(self, anno):
        anno = anno.copy()

        if self.scale is not None:
            anno.x_top_left *= self.scale
            anno.y_top_left *= self.scale
            anno.width *= self.scale
            anno.height *= self.scale
        if self.pad is not None:
            anno.x_top_left += self.pad[0]
            anno.y_top_left += self.pad[1]

        return anno


class Pad(BaseMultiTransform):
    """ Pad images/annotations to a certain dimension.

    Args:
        dimension (int or tuple, optional): Default size for the padding, expressed as a single integer or as a (width, height) tuple; Default **None**
        dataset (lightnet.data.Dataset, optional): Dataset that uses this transform; Default **None**
        multiple_dim (boolean, optional): Consider given dimensions to be multiples instead of exact values; Default **True**
        fill_color (int or float, optional): Fill color to be used for padding (if int, will be divided by 255); Default **0.5**

    Warning:
        Do note that the ``dimension`` or ``dataset`` argument here uses the given width and height as a multiple instead of a real dimension by default.
        Given a certain value X, the image (and annotations) will be padded, so that the image dimensions are a multiple of X. |br|
        This is different compared to :class:`~lightnet.data.transform.Crop` or :class:`~lightnet.data.transform.Letterbox`.

        You can toggle this behaviour by setting ``multiple_dim=False``, but keep in mind that the given dimensions should always be bigger than the original input image dimensions.

    Note:
        Create 1 Pad object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.
    """
    def __init__(self, dimension=None, dataset=None, multiple_dim=True, fill_color=127):
        super().__init__()
        self.dimension = dimension
        self.dataset = dataset
        self.multiple_dim = multiple_dim
        self.fill_color = fill_color if isinstance(fill_color, float) else fill_color / 255
        if self.dimension is None and self.dataset is None:
            raise ValueError('This transform either requires a dimension or a dataset to infer the dimension')

        self.pad = None
        self.scale = None

    def _get_params(self, im_w, im_h):
        if self.dataset is not None:
            net_w, net_h = self.dataset.input_dim
        elif isinstance(self.dimension, int):
            net_w, net_h = self.dimension, self.dimension
        else:
            net_w, net_h = self.dimension

        if self.multiple_dim:
            if im_w % net_w == 0 and im_h % net_h == 0:
                self.pad = None
            else:
                pad_w = ((net_w - (im_w % net_w)) % net_w) / 2
                pad_h = ((net_h - (im_h % net_h)) % net_h) / 2
                self.pad = (int(pad_w), int(pad_h), int(pad_w+.5), int(pad_h+.5))
        else:
            if im_w == net_w and im_h == net_h:
                self.pad = None
            elif im_w <= net_w and im_h <= net_h:
                pad_w = (net_w - im_w) / 2
                pad_h = (net_h - im_h) / 2
                self.pad = (int(pad_w), int(pad_h), int(pad_w+.5), int(pad_h+.5))
            else:
                raise ValueError(f'Can only pad to bigger dimensions. Image is bigger than network dimensions [({im_h}, {im_w}) -> ({net_h}, {net_w})]')

    def _tf_pil(self, img):
        im_w, im_h = img.size
        self._get_params(im_w, im_h)

        # Pad
        if self.pad is not None:
            shape = np.array(img).shape
            channels = shape[2] if len(shape) > 2 else 1
            img = ImageOps.expand(img, border=self.pad, fill=(int(self.fill_color*255),)*channels)

        return img

    def _tf_cv(self, img):
        im_h, im_w = img.shape[:2]
        self._get_params(im_w, im_h)

        # Pad
        if self.pad is not None:
            channels = img.shape[2] if len(img.shape) > 2 else 1
            img = cv2.copyMakeBorder(img, self.pad[1], self.pad[3], self.pad[0], self.pad[2], cv2.BORDER_CONSTANT, value=(int(self.fill_color*255),)*channels)

        return img

    def _tf_torch(self, img):
        im_h, im_w = img.shape[-2:]
        self._get_params(im_w, im_h)

        # Pad
        if self.pad is not None:
            img = torch.nn.functional.pad(img, (self.pad[0], self.pad[2], self.pad[1], self.pad[3]), value=self.fill_color)

        return img

    def _tf_anno(self, anno):
        anno = anno.copy()

        if self.pad is not None:
            anno.x_top_left += self.pad[0]
            anno.y_top_left += self.pad[1]

        return anno


class FitAnno(BaseMultiTransform):
    """ Crop and filter annotations to fit inside of the image boundaries. |br|
    This transformation also modifies the `truncated` columns of the annotations, by computing how much of the annotation was cut off.

    Args:
        crop (boolean, optional): Whether to actually crop annotations to fit inside the image boundaries; Default **True**
        filter (boolean, optional): Whether to filter the annotations if they are not completely inside of the image boundaries; Default **True**
        filter_type (string, optional): How to filter ('remove' or 'ignore'); Default **remove**
        filter_threshold (number, optional): Minimal percentage of the bounding box area that still needs to be inside the image; Default **0.001**
        remove_empty (boolean, optional): Whether to remove annotations whose bounding box area is zero (independent of filter args); Default **True**

    Note:
        This transformation does not modify the image data, but is still a multi-transform as it needs to read the image to get the dimensions.
        Create 1 FitAnno object and use it for both image and annotation transforms.

    Note:
        If the `filter_threshold` is a tuple of 2 numbers, then they are to be considered as **(width, height)** threshold values.
        Ohterwise the threshold is to be considered as an area threshold.
    """
    def __init__(self, crop=True, filter=True, filter_type='remove', filter_threshold=0.001, remove_empty=True):
        super().__init__()
        self.crop = crop
        self.filter = filter
        self.filter_type = filter_type.lower()
        self.filter_threshold = filter_threshold
        self.remove_empty = True

        self.image_dim = None

    def _get_params(self, im_w, im_h):
        self.image_dim = (im_w, im_h)

    def _tf_pil(self, img):
        im_w, im_h = img.size
        self._get_params(im_w, im_h)
        return img

    def _tf_cv(self, img):
        im_h, im_w = img.shape[:2]
        self._get_params(im_w, im_h)
        return img

    def _tf_torch(self, img):
        im_h, im_w = img.shape[-2:]
        self._get_params(im_w, im_h)
        return img

    def _tf_anno(self, anno):
        anno = anno.copy()

        crop_coords = np.empty((4, len(anno.index)), dtype=np.float64)
        crop_coords[0] = anno.x_top_left.values
        crop_coords[1] = crop_coords[0] + anno.width.values
        crop_coords[2] = anno.y_top_left.values
        crop_coords[3] = crop_coords[2] + anno.height.values
        crop_coords[:2] = crop_coords[:2].clip(0, self.image_dim[0])
        crop_coords[2:] = crop_coords[2:].clip(0, self.image_dim[1])
        crop_width = crop_coords[1] - crop_coords[0]
        crop_height = crop_coords[3] - crop_coords[2]

        # UserWarnings occur when box width or height is zero (divide by zero)
        # Disable theses annoying warnings as we manually handle the nan cases:
        #   - Masks: `nan >= X = False`
        #   - Computes: `np.nan_to_num(nan) = 0`
        with np.errstate(divide='ignore', invalid='ignore'):
            # Filter
            if self.filter:
                if isinstance(self.filter_threshold, collections.Sequence):
                    mask = (
                        ((crop_width / anno.width.values) >= self.filter_threshold[0])
                        & ((crop_height / anno.height.values) >= self.filter_threshold[1])
                    )
                else:
                    mask = ((crop_width * crop_height) / (anno.width.values * anno.height.values)) >= self.filter_threshold

                if self.filter_type == 'ignore':
                    anno.loc[~mask, 'ignore'] = True
                else:
                    anno = anno[mask].copy()
                    if len(anno.index) == 0:
                        return anno

                    crop_coords = crop_coords[:, mask]
                    crop_width = crop_width[mask]
                    crop_height = crop_height[mask]

            # Crop
            if self.crop:
                anno.truncated = np.nan_to_num((1 - ((crop_width * crop_height * (1 - anno.truncated.values)) / (anno.width.values * anno.height.values))).clip(0, 1))
                anno.x_top_left = crop_coords[0]
                anno.y_top_left = crop_coords[2]
                anno.width = crop_width
                anno.height = crop_height

        # Remove empty
        if self.remove_empty:
            anno = anno[(anno.width > 0) & (anno.height > 0)].copy()

        return anno



#
#   Combine multiple pruners
#   Copyright EAVISE
#
import collections
import logging

__all__ = ['MultiPruner']
log = logging.getLogger(__name__)


def flatten(x):
    """ Only works with lists of non-iterable data (eg. not strings) """
    if isinstance(x, collections.Iterable):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]


class MultiPruner:
    """ Combines different pruner methods together. |br|
    This class takes a list of pruners and applies each of them in order.
    It can be used to simultaneously prune different networks or different parts of a network,
    or even run different pruners on a single network.

    Args:
        *pruners (list <lightnet.prune.Pruner>): List of pruners to run

    Note:
        The MultiPruner runs pruners in the order they were declared upon initialisation.
    """
    def __init__(self, *pruners):
        self.pruners = flatten(pruners)

    def __call__(self, *percentages):
        """ Actually perform the pruning.

        Args:
            *percentages (list <float>): percentage for each pruner

        Note:
            If the percentage for a certain pruner is zero, that pruner gets skipped.
        """
        percentages = flatten(percentages)
        if len(percentages) < len(self.pruners):
            raise ValueError(f'Received {len(percentages)} percentages, but we have {len(self.pruners)} pruners.')
        elif len(percentages) > len(self.pruners):
            log.error(f'Received {len(percentages)} percentages, but we have {len(self.pruners)} pruners.')

        total_pruned = 0
        for pruner, percentage in zip(self.pruners, percentages):
            if percentage <= 0:
                continue
            total_pruned += pruner(percentage)

        return total_pruned

    @property
    def prunable_channels(self):
        """ Returns the number of prunable channels left on the module. |br|
        This property is added, so that you can use a :class:`~lightnet.prune.MultiPruner`
        anywhere you would use a regular :class:`~lightnet.prune.Pruner`,
        but it does not make a whole lot of sense for this pruner, as each individual pruner could have a different model.

        We thuse simply return the ``prunable_channels`` property from the first pruner.

        Returns:
            int: Number of prunable channels of the entire model
        """
        return self.pruners[0].prunable_channels

    @property
    def pruned_channels(self):
        """ Returns the number of pruned channels from the last pruning operation. """
        return sum(p.pruned_channels for p in self.pruners)

    @property
    def soft_pruned_channels(self):
        """ Returns the number of soft pruned channels from the last pruning operation. """
        return sum(p.soft_pruned_channels for p in self.pruners)

    @property
    def hard_pruned_channels(self):
        """ Returns the number of hard pruned channels from the last pruning operation. """
        return sum(p.hard_pruned_channels for p in self.pruners)

    def __str__(self):
        string = f'{self.__class__.__name__} ['
        for pruner in self.pruners:
            string += f'{str(pruner)}, '
        return string[:-2] + ']'

    def __repr__(self):
        string = f'{self.__class__.__name__} [\n'
        for pruner in self.pruners:
            prunerrepr = repr(pruner)
            if '\n' in prunerrepr:
                prunerrepr = prunerrepr.replace('\n', '\n    ')
            string += f'  {prunerrepr},\n'
        return string + ']'



#
#   Darknet Tiny YOLOv3 model
#   Copyright EAVISE
#

import functools
from collections import OrderedDict, Iterable
import torch
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['TinyYoloV3']


class TinyYoloV3(lnn.module.Darknet):
    """ Tiny Yolo v3 implementation :cite:`yolo_v3`.

    Args:
        num_classes (Number, optional): Number of classes; Default **80**
        input_channels (Number, optional): Number of input channels; Default **3**
        anchors (list, optional): 3D list with anchor values; Default **Tiny Yolo v3 anchors (COCO)**

    Attributes:
        self.stride: Subsampling factors of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
        self.remap_darknet: Remapping rules for weights from the `~lightnet.models.Darknet` model.

    Note:
        Unlike YoloV2, the anchors here are defined as multiples of the input dimensions and not as a multiple of the output dimensions!
        The anchor list also has one more dimension than the one from YoloV2, in order to differentiate which anchors belong to which stride.

    Warning:
        The :class:`~lightnet.network.loss.MultiScaleRegionLoss` and :class:`~lightnet.data.transform.GetMultiScaleBoundingBoxes`
        do not implement the overlapping class labels of the original implementation.
        Your weight files from darknet will thus not have the same accuracies as in darknet itself.
    """
    stride = (32, 16)
    inner_stride = 32
    remap_darknet = [
        (r'^layers.0.(\d+_)',   r'extractor.\1'),  # All base layers (1-13)
    ]

    def __init__(self, num_classes, input_channels=3, anchors=[[(81, 82), (135, 169), (344, 319)], [(10, 14), (23, 27), (37, 58)]]):
        super().__init__()
        if not isinstance(anchors, Iterable) and not isinstance(anchors[0], Iterable) and not isinstance(anchors[0][0], Iterable):
            raise TypeError('Anchors need to be a 3D list of numbers')

        # Parameters
        self.num_classes = num_classes
        self.input_channels = input_channels
        self.anchors = []   # YoloV3 defines anchors as a multiple of the input dimensions of the network as opposed to the output dimensions
        for i, s in enumerate(self.stride):
            self.anchors.append([(a[0] / s, a[1] / s) for a in anchors[i]])

        # Network
        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01
        self.extractor = lnn.layer.SequentialSelect(
            ['9_convbatch'], True,
            # Sequence 0 : input = input_channels
            OrderedDict([
                ('1_convbatch',     lnn.layer.Conv2dBatchReLU(input_channels, 16, 3, 1, 1, relu=relu, momentum=momentum)),
                ('2_maxpool',       nn.MaxPool2d(2, 2)),
                ('3_convbatch',     lnn.layer.Conv2dBatchReLU(16, 32, 3, 1, 1, relu=relu, momentum=momentum)),
                ('4_maxpool',       nn.MaxPool2d(2, 2)),
                ('5_convbatch',     lnn.layer.Conv2dBatchReLU(32, 64, 3, 1, 1, relu=relu, momentum=momentum)),
                ('6_maxpool',       nn.MaxPool2d(2, 2)),
                ('7_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ('8_maxpool',       nn.MaxPool2d(2, 2)),
                ('9_convbatch',     lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ('10_maxpool',      nn.MaxPool2d(2, 2)),
                ('11_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('12_maxpool',      lnn.layer.PaddedMaxPool2d(2, 1, (0, 1, 0, 1))),
                ('13_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('14_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 256, 1, 1, 0, relu=relu, momentum=momentum)),
            ]),
        )

        self.detector = nn.ModuleList([
            # Sequence 0 : input = extractor
            nn.Sequential(
                OrderedDict([
                    ('15_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('16_conv',         nn.Conv2d(512, len(self.anchors[0])*(5+self.num_classes), 1, 1, 0)),
                ])
            ),

            # Sequence 1 : input = extractor
            nn.Sequential(
                OrderedDict([
                    ('17_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('18_upsample',     nn.Upsample(scale_factor=2, mode='nearest')),
                ])
            ),

            # Sequence 2 : input = 18_upsample + 9_convbatch
            nn.Sequential(
                OrderedDict([
                    ('19_convbatch',    lnn.layer.Conv2dBatchReLU(128+256, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('20_conv',         nn.Conv2d(256, len(self.anchors[1])*(5+self.num_classes), 1, 1, 0)),
                ])
            ),
        ])

    def forward(self, x):
        # Feature extractor
        x, inter_features = self.extractor(x)

        # Detector 0
        out_0 = self.detector[0](x)

        # Detector 1
        x = self.detector[1](x)
        out_1 = self.detector[2](torch.cat((x, inter_features), 1))

        return (out_0, out_1)



#
#   Fusion module
#   Copyright EAVISE
#

import copy
import logging
from collections import OrderedDict
import torch
import torch.nn as nn
from torch.nn.modules.module import _addindent


__all__ = ['Fusion']
log = logging.getLogger(__name__)


class Fusion(nn.Module):
    """ Fusion module from :cite:`rgbd_fusion_v2`.
    This module is like a :class:`torch.nn.Sequential`, but it will perform the actions twice,
    once on the regular input and once on the fusion input. |br|
    The fusion will be performed by adding an extra 1x1 fuse convolution between the output of both streams and the input of the combined stream,
    to mix both streams and reduce the number of output feature maps by a factor of 2. |br|
    This module takes a single input feature map during its forward pass, and splits it evenly for both input streams.

    Args:
        layers (dict or list of pytorch modules): Layers that will be used. These layers are internally passed to a :class:`~torch.nn.Sequential` and must thus comply with the rules for this class
        fuse_layer (int, optional): Number between 0 and the number of layers + 1, that controls where to fuse both streams; Default **None**

    .. figure:: /.static/api/fusion.*
       :width: 100%
       :alt: Fusion module and fuse layer

    Note:
        Depending on the value of the `fuse_layer` attribute, fusion is performed at different stages of the module. |br|

        - If no `fuse\\_layer` is given (or **None** is given as its value),\
        no fusion will be done and the input will be considered as an already fused combination.
        - If the `fuse\\_layer` attribute is an integer from 0 to :math:`num\\_layers`, the module will fuse both streams after the number of the layer that is given.\
        Giving a value of **0** thus means to fuse before the first layer and giving a value of **num_layers** to fuse after the last.
        - Finally, if :math:`fuse\\_layer == num\\_layers + 1`, then no fusion will occur, but rather both streams will be processed seperately\
        and the output feature maps will simply be concatenated at the end.

        These rules allow the chain multiple :class:`~lightnet.network.layer.Fusion` modules together, only fusing in one of them at a certain time.

    Example:
        >>> layers = [
        ...   ln.network.layer.Conv2dBatchReLU(3,  32, 3, 1, 1),
        ...   ln.network.layer.Conv2dBatchReLU(32, 64, 3, 1, 1),
        ...   ln.network.layer.Conv2dBatchReLU(64, 32, 3, 1, 1),
        ... ]
        >>> # Streams are fused in the middle of the module
        >>> module = ln.network.layer.Fusion(layers, fuse_layer=1)
        >>> print(module)
        Fusion(
          (Regular & Fusion): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
          (Fuse): Conv2d(64, 32, kernel_size=(1, 1), stride=(1, 1), bias=False)
          (Combined): Sequential(
            (0): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
        )
        >>> # Streams are fused at the end of the module (after last convolution)
        >>> module = ln.network.layer.Fusion(layers, fuse_layer=3)
        >>> print(module)
        Fusion(
          (Regular & Fusion): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
          (Fuse): Conv2d(64, 32, kernel_size=(1, 1), stride=(1, 1), bias=False)
        )
        >>> # Streams are fused before the first layer
        >>> module = ln.network.layer.Fusion(layers, fuse_layer=0)
        >>> print(module)
        Fusion(
          (Fuse): Conv2d(6, 3, kernel_size=(1, 1), stride=(1, 1), bias=False)
          (Combined): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
        )
        >>> # Streams were fused before this module and thus there is no fusion involved (only combined sequential)
        >>> module = ln.network.layer.Fusion(layers, fuse_layer=None)
        >>> print(module)
        Fusion(
          (Combined): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
        )
        >>> # Streams are not fused in this module (duplicated regular and fusion sequentials)
        >>> module = ln.network.layer.Fusion(layers, fuse_layer=4)
        >>> print(module)
        Fusion(
          (Regular & Fusion): Sequential(
            (0): Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (1): Conv2dBatchReLU(32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
            (2): Conv2dBatchReLU(64, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
          )
        )

    Warning:
        This module will create a :class:`~torch.nn.Sequential` for the regular convolutional stream and deepcopy that for the fusion stream.
        This will effectively create 2 different streams that have their own weights, but it does mean that both streams start with identical weights. |br|
        It is strongly advised to use pretrained weights or initialize your weights randomly after having created these modules.

    Warning:
        The way we compute the input and output feature maps for the 1x1 fuse convolution,
        is by looping through the regular stream or combined stream,
        looking for the last `out_channels` or first `in_channels` attribute of the layers respectively. |br|
        This means that this module only works if there are convolutional layers in the list,
        or any other layer that has these `in_channels` and `out_channels` attributes to be able to deduct the number of feature maps.
    """
    def __init__(self, layers, fuse_layer=None):
        super().__init__()

        # Parameters
        self.fuse_layer = fuse_layer

        # layers
        if self.fuse_layer is None:             # Combined
            self.regular = None
            self.fusion = None
            if isinstance(layers, dict):
                self.combined = nn.Sequential(layers)
            else:
                self.combined = nn.Sequential(*layers)
            self.fuse = None
        elif self.fuse_layer == 0:              # Fuse + Combined
            self.regular = None
            self.fusion = None
            if isinstance(layers, dict):
                self.combined = nn.Sequential(layers)
            else:
                self.combined = nn.Sequential(*layers)
            self.fuse = self._get_fuse_conv()
        elif self.fuse_layer == len(layers):    # Reg/Fusion + Fuse
            if isinstance(layers, dict):
                self.regular = nn.Sequential(layers)
            else:
                self.regular = nn.Sequential(*layers)
            self.fusion = copy.deepcopy(self.regular)
            self.combined = None
            self.fuse = self._get_fuse_conv()
        elif self.fuse_layer > len(layers):     # Reg/Fusion
            if self.fuse_layer > len(layers) + 1:
                log.debug(f'fuse_layer variable is too high, setting it to {len(layers)+1} which will not perform any fusion [{self.fuse_layer}/{len(layers)+1}]')
                self.fuse_layer = len(layers) + 1

            if isinstance(layers, dict):
                self.regular = nn.Sequential(layers)
            else:
                self.regular = nn.Sequential(*layers)
            self.fusion = copy.deepcopy(self.regular)
            self.combined = None
            self.fuse = None
        elif self.fuse_layer < len(layers):     # Reg/Fusion + Fuse + Combined
            if isinstance(layers, dict):
                self.regular = nn.Sequential(OrderedDict(list(layers.items())[:self.fuse_layer]))
                self.fusion = copy.deepcopy(self.regular)
                self.combined = nn.Sequential(OrderedDict(list(layers.items())[self.fuse_layer:]))
            else:
                self.regular = nn.Sequential(*layers[:self.fuse_layer])
                self.fusion = copy.deepcopy(self.regular)
                self.combined = nn.Sequential(*layers[self.fuse_layer:])
            self.fuse = self._get_fuse_conv()

    def __repr__(self):
        main_str = self._get_name() + '('

        if self.regular is not None:
            mod_str = _addindent(repr(self.regular), 2)
            main_str += '\n  (Regular & Fusion): ' + mod_str
        if self.fuse is not None:
            mod_str = _addindent(repr(self.fuse), 2)
            main_str += '\n  (Fuse): ' + mod_str
        if self.combined is not None:
            mod_str = _addindent(repr(self.combined), 2)
            main_str += '\n  (Combined): ' + mod_str

        return main_str + '\n)'

    def _get_fuse_conv(self):
        channels = None
        if self.combined is not None:
            channels = find_attr(self.combined, 'in_channels')
        if channels is None and self.regular is not None:
            channels = find_attr(self.regular, 'out_channels', first=False)
        if channels is None:
            raise TypeError('Could not find "in_channels" or "out_channels" attribute in layers.')

        return nn.Conv2d(channels*2, channels, 1, 1, 0, bias=False).to(list(self.parameters())[0].device)

    def forward(self, x):
        if self.regular is not None:
            channels = x.size(1)
            if channels % 2 != 0:
                raise ValueError(f'Number of input channels is not divisible by 2 [{channels}]')

            r = self.regular(x[:, :channels//2])
            f = self.fusion(x[:, channels//2:])
            x = torch.cat((r, f), 1)

        if self.fuse is not None:
            x = self.fuse(x)

        if self.combined is not None:
            x = self.combined(x)

        return x


def find_attr(module, name, first=True):
    if hasattr(module, name):
        return getattr(module, name)

    retval = None
    for mod in module.children():
        r = find_attr(mod, name, first)
        if r is not None:
            if first:
                return r
            else:
                retval = r

    return retval



#
#   Lightnet dataset that uses the same files and structure as darknet and performs the same data augmentations.
#   Copyright EAVISE
#

import os
from PIL import Image
from torchvision import transforms as tf
import lightnet.data as lnd
from . import BramboxDataset

try:
    import brambox as bb
except ImportError:
    bb = None


__all__ = ['DarknetDataset']


class DarknetDataset(BramboxDataset):
    """ Dataset that works with darknet files and performs the same data augmentations.
    You must use this dataset with the :meth:`~lightnet.data.brambox_collate` function in a dataloader.
    If you enable the data augmentation you must also use the :class:`~lightnet.data.DataLoader` class as dataloader.

    Args:
        data_file (str): File containing path to image files (relative from where command is run)
        class_label_map (list): class label map to convert darknet class indices to a name
        augment (Boolean, optional): Whether or not you want data augmentation; Default **True**
        input_dimension (tuple): Input dimension of the network width,height; Default **416,416**
        jitter (Number [0-1], optional): Determines random crop sizes; Default **0.2**
        flip (Number [0-1], optional): Determines whether image will be flipped; Default **0.5**
        hue (Number, optional): Determines hue shift; Default **0.1**
        saturation (Number, optional): Determines saturation shift; Default **1.5**
        value (Number, optional): Determines value (exposure) shift; Default **1.5**

    Returns:
        tuple: image_tensor, list of brambox boxes
    """
    def __init__(self, data_file, class_label_map, augment=True, input_dimension=(416, 416), jitter=.3, flip=.5, hue=.1, saturation=1.5, value=1.5):
        if bb is None:
            raise ImportError('Brambox needs to be installed to use this dataset')

        def identify(name):
            return self.img_paths[self.anno_paths.get_loc(name)]

        def get_image_dimensions(name):
            with Image.open(identify(name)) as img:
                return img.size

        # Get paths
        with open(data_file, 'r') as f:
            self.img_paths = f.read().splitlines()
        self.anno_paths = [os.path.splitext(p)[0]+'.txt' for p in self.img_paths]

        # Load data
        annos = bb.io.load(
            'anno_darknet',
            self.anno_paths,
            identify=lambda f: f,
            class_label_map=class_label_map,
            image_dims=get_image_dimensions,
        )

        # Data transformation
        lb = lnd.transform.Letterbox(dataset=self)
        rf = lnd.transform.RandomFlip(flip)
        rc = lnd.transform.RandomJitter(jitter, True)
        hsv = lnd.transform.RandomHSV(hue, saturation, value)
        it = tf.ToTensor()
        if augment:
            img_tf = lnd.transform.Compose([hsv, rc, rf, lb, it])
            anno_tf = lnd.transform.Compose([rc, rf, lb])
        else:
            img_tf = lnd.transform.Compose([lb, it])
            anno_tf = lnd.transform.Compose([lb])

        super().__init__(annos, input_dimension, class_label_map, identify, img_tf, anno_tf)



#
#   Darknet related layers
#   Copyright EAVISE
#

import logging
import torch
import torch.nn as nn
import torch.nn.functional as F


__all__ = ['Conv2dBatchReLU', 'Flatten', 'PaddedMaxPool2d', 'Reorg']
log = logging.getLogger(__name__)


class Conv2dBatchReLU(nn.Module):
    """ This convenience layer groups a 2D convolution, a batchnorm and a ReLU.
    They are executed in a sequential manner.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int or tuple): Size of the kernel of the convolution
        stride (int or tuple): Stride of the convolution
        padding (int or tuple): padding of the convolution
        momentum (number, optional): momentum of the moving averages of the normalization; Default **0.1**
        relu (class, optional): Which ReLU to use; Default :class:`torch.nn.ReLU`

    .. figure:: /.static/api/conv2dbatchrelu.*
       :width: 100%
       :alt: Conv2dBatchReLU module design

    Note:
        The bias term in the :class:`~torch.nn.Conv2d` is disabled for this module.

    Example:
        >>> module = ln.network.layer.Conv2dBatchReLU(3, 32, 3, 1, 1)
        >>> print(module)
        Conv2dBatchReLU(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 10, 10])
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding, momentum=0.1, relu=lambda: nn.ReLU(inplace=True)):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, bias=False),
            nn.BatchNorm2d(out_channels, momentum=momentum),
            relu()
        )

    def __repr__(self):
        s = '{name}({in_channels}, {out_channels}, kernel_size={kernel_size}, stride={stride}, padding={padding}, {relu})'
        return s.format(
            name=self.__class__.__name__,
            in_channels=self.layers[0].in_channels,
            out_channels=self.layers[0].out_channels,
            kernel_size=self.layers[0].kernel_size,
            stride=self.layers[0].stride,
            padding=self.layers[0].padding,
            relu=self.layers[2],
        )

    def forward(self, x):
        x = self.layers(x)
        return x


class Flatten(nn.Module):
    """ Flatten tensor into single dimension.

    Args:
        batch (boolean, optional): If True, consider input to be batched and do not flatten first dim; Default **True**

    Example:
        >>> # By default batch_mode is true
        >>> module = ln.network.layer.Flatten()
        >>> in_tensor = torch.rand(8, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([8, 300])

        >>> # Disable batch_mode
        >>> module = ln.network.layer.Flatten(False)
        >>> in_tensor = torch.rand(8, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([2400])
    """
    def __init__(self, batch=True):
        super().__init__()
        self.batch = batch

    def forward(self, x):
        if self.batch:
            return x.view(x.size(0), -1)
        else:
            return x.view(-1)


class PaddedMaxPool2d(nn.Module):
    """ Maxpool layer with replicate-padding instead of the zero-padding from :class:`torch.nn.MaxPool2d`. |br|
    This layer is not a traditional pooling layer in the sence that it does not modify the dimension of the input tensor.

    Args:
        kernel_size (int or tuple): Kernel size for maxpooling
        stride (int or tuple, optional): The stride of the window; Default ``kernel_size``
        padding (tuple, optional): (left, right, top, bottom) padding; Default **None**
        dilation (int or tuple, optional): A parameter that controls the stride of elements in the window

    Example:
        >>> module = ln.network.layer.PaddedMaxPool2d(2, 1, (0, 1, 0, 1))
        >>> print(module)
        PaddedMaxPool2d(kernel_size=2, stride=1, padding=(0, 1, 0, 1), dilation=1)
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 3, 10, 10])
    """
    def __init__(self, kernel_size, stride=None, padding=(0, 0, 0, 0), dilation=1):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride or kernel_size
        self.padding = padding
        self.dilation = dilation

    def extra_repr(self):
        return f'kernel_size={self.kernel_size}, stride={self.stride}, padding={self.padding}, dilation={self.dilation}'

    def forward(self, x):
        x = F.max_pool2d(F.pad(x, self.padding, mode='replicate'), self.kernel_size, self.stride, 0, self.dilation)
        return x


class Reorg(nn.Module):
    """ This layer reorganizes a tensor according to a stride.
    The width and height dimensions (2 and 3) will be sliced by the stride and then stacked in dimension 1. (input must have 4 dimensions)

    Args:
        stride (int): stride to divide the input tensor

    Note:
        This implementation follows the darknet reorg layer implementation, which we took from this `issue <reorglink_>`_. |br|
        This specific implementation requires that the channel dimension should be divisible by :math:`stride^{\,2}`.

    Example:
        >>> # Divide width and height by 2 and stack in channels (thus multiplying by 4)
        >>> module = ln.network.layer.Reorg(stride=2)
        >>> in_tensor = torch.rand(8, 4, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([8, 16, 5, 5])

        >>> # Divide width and height by 4, Note that channel should thus be divisible by 4^2
        >>> module = ln.network.layer.Reorg(stride=4)
        >>> in_tensor = torch.rand(8, 16, 16, 16)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([8, 256, 4, 4])

    .. _reorglink: https://github.com/thtrieu/darkflow/issues/173#issuecomment-296048648
    """
    def __init__(self, stride=2):
        super().__init__()
        self.stride = stride
        if not isinstance(stride, int):
            raise TypeError(f'stride is not an int [{type(stride)}]')

    def extra_repr(self):
        return f'stride={self.stride}'

    def forward(self, x):
        assert x.dim() == 4
        B, C, H, W = x.size()
        assert H % self.stride == 0, f'Dimension height mismatch: {H} is not divisible by {self.stride}'
        assert W % self.stride == 0, f'Dimension width mismatch: {W} is not divisible by {self.stride}'
        mem_fmt = x.is_contiguous(memory_format=torch.channels_last)

        x = x.reshape(B, C//(self.stride**2), H, self.stride, W, self.stride)
        x = x.permute(0, 3, 5, 1, 2, 4)
        x = x.reshape(B, -1, H//self.stride, W//self.stride)
        if mem_fmt:
            x = x.contiguous(memory_format=torch.channels_last)

        return x



#
#   Darknet YoloV3 RegionLoss
#   Copyright EAVISE
#

import torch
from . import RegionLoss

__all__ = ['MultiScaleRegionLoss']


class MultiScaleRegionLoss(RegionLoss):
    """ Computes region loss from darknet network output and target annotation at multiple scales (yoloV3).

    Args:
        num_classes (int): number of classes to detect
        anchors (list): 3D list representing multiple 2D lists with anchor boxes for the different scales
        stride (list): The different downsampling factors of the different network outputs
        seen (optional, torch.Tensor): How many images the network has already been trained on; Default **0**
        coord_scale (optional, float): weight of bounding box coordinates; Default **1.0**
        noobject_scale (optional, float): weight of regions without target boxes; Default **1.0**
        object_scale (optional, float): weight of regions with target boxes; Default **5.0**
        class_scale (optional, float): weight of categorical predictions; Default **1.0**
        thresh (optional, float): minimum iou between a predicted box and ground truth for them to be considered matching; Default **0.6**
        coord_prefill (optional, int): This parameter controls for how many images the network will prefill the target coordinates, biassing the network to predict the center at **.5,.5**; Default **12800**

    Note:
        All parameters are the same as :class:`~lightnet.network.loss.RegionLoss`, except for `anchors` and `stride`. |br|
        These 2 parameters need separate values for each different network output scale and thus need to be lists of the original parameter.

    Warning:
        This loss function is not entirely equivalent to the Darknet implementation! |br|
        We just execute the regular :class:`~lightnet.network.loss.RegionLoss` at multiple scales (different strides and anchors),
        and as such did not implement overlapping class labels.
    """
    def __init__(self, num_classes, anchors, stride, **kwargs):
        super().__init__(num_classes, anchors[0], stride=stride[0], **kwargs)

        if len(anchors) != len(stride):
            raise IndexError('length of anchors and stride should be equal (number of output scales)')
        self._anchors = torch.tensor(anchors, dtype=torch.float, requires_grad=False)
        self._stride = stride

    def extra_repr(self):
        repr_str = f'classes={self.num_classes}, stride={self.stride}, threshold={self.thresh}, seen={self.seen.item()}\n'
        repr_str += f'coord_scale={self.coord_scale}, object_scale={self.object_scale}, noobject_scale={self.noobject_scale}, class_scale={self.class_scale}\n'
        repr_str += f'anchors='
        start = True
        for anchors in self._anchors:
            if not start:
                repr_str += '| '
            for a in anchors:
                repr_str += f'[{a[0]:.5g}, {a[1]:.5g}] '
            start = False
        return repr_str

    def forward(self, output, target, seen=None):
        device = output[0].device
        loss_total = torch.tensor(0.0, device=device)
        loss_conf = torch.tensor(0.0, device=device)
        loss_coord = torch.tensor(0.0, device=device)
        loss_class = torch.tensor(0.0, device=device)
        if seen is not None:
            self.seen = torch.tensor(seen)

        # Run loss at different scales and sum resulting loss values
        for i, out in enumerate(output):
            self.anchors = self._anchors[i]
            self.num_anchors = self.anchors.shape[0]
            self.anchor_step = self.anchors.shape[1]
            self.stride = self._stride[i]

            super().forward(out, target)
            loss_total += self.loss_total
            loss_conf += self.loss_conf
            loss_coord += self.loss_coord
            loss_class += self.loss_class

        # Overwrite loss values with avg
        self.loss_total = loss_total / len(output)
        self.loss_conf = loss_conf / len(output)
        self.loss_coord = loss_coord / len(output)
        self.loss_class = loss_class / len(output)
        return self.loss_total



#
#   Darknet Tiny YOLOv2 model
#   Copyright EAVISE
#

import functools
import logging
from collections import OrderedDict, Iterable
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['TinyYoloV2', 'TinyYolo']
log = logging.getLogger('lightnet.models')


class TinyYoloV2(lnn.module.Darknet):
    """ Tiny Yolo v2 implementation :cite:`yolo_v2`.

    Args:
        num_classes (Number, optional): Number of classes; Default **20**
        input_channels (Number, optional): Number of input channels; Default **3**
        anchors (list, optional): 2D list with anchor values; Default **Tiny yolo v2 anchors (VOC)**

    Attributes:
        self.stride: Subsampling factor of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
        self.remap_darknet: Remapping rules for weights from the :class:`~lightnet.models.Darknet` model.
    """
    stride = 32
    inner_stride = 32
    remap_darknet = [
        (r'^layers.0.(\d+_)',   r'layers.\1'),  # All base layers (1-13)
    ]

    def __init__(self, num_classes, input_channels=3, anchors=[(1.08, 1.19), (3.42, 4.41), (6.63, 11.38), (9.42, 5.11), (16.62, 10.52)]):
        super().__init__()
        if not isinstance(anchors, Iterable) and not isinstance(anchors[0], Iterable):
            raise TypeError('Anchors need to be a 2D list of numbers')

        # Parameters
        self.num_classes = num_classes
        self.input_channels = input_channels
        self.anchors = anchors

        # Network
        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01
        self.layers = nn.Sequential(
            OrderedDict([
                ('1_convbatch',     lnn.layer.Conv2dBatchReLU(input_channels, 16, 3, 1, 1, relu=relu, momentum=momentum)),
                ('2_max',           nn.MaxPool2d(2, 2)),
                ('3_convbatch',     lnn.layer.Conv2dBatchReLU(16, 32, 3, 1, 1, relu=relu, momentum=momentum)),
                ('4_max',           nn.MaxPool2d(2, 2)),
                ('5_convbatch',     lnn.layer.Conv2dBatchReLU(32, 64, 3, 1, 1, relu=relu, momentum=momentum)),
                ('6_max',           nn.MaxPool2d(2, 2)),
                ('7_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ('8_max',           nn.MaxPool2d(2, 2)),
                ('9_convbatch',     lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ('10_max',          nn.MaxPool2d(2, 2)),
                ('11_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('12_max',          lnn.layer.PaddedMaxPool2d(2, 1, (0, 1, 0, 1))),
                ('13_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('14_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('15_conv',         nn.Conv2d(1024, len(self.anchors)*(5+self.num_classes), 1, 1, 0)),
            ])
        )


def TinyYolo(*args, **kwargs):
    """ DEPRECATED : Please use the more aptly named TinyYoloV2 class. """
    log.deprecated('This class is deprecated! Please use the more aptly named TinyYoloV2 class.')
    return TinyYoloV2(*args, **kwargs)



#
#   Tree data structure and related functionality
#   Copyright EAVISE
#
__all__ = ['Node', 'traverse_depth_first', 'traverse_breadth_first']


class Node:
    """ Tree node.
    This data structure is used to get the dependency tree when pruning a convolution.
    """
    def __init__(self, nodetype, nodename=None, module=None, metadata=None, parents=None, children=None):
        self.type = nodetype
        self.name = nodename
        self.module = module
        self.metadata = metadata
        self.parents = parents if parents is not None else []
        self.children = children if children is not None else []

    def __str__(self):
        return f'Node({self.type.name}, {self.name}, #parents={len(self.parents)}, #children={len(self.children)})'

    def __repr__(self):
        string = f'Node({self.type.name}, {self.name}, parents=['
        string += ', '.join(str(p) for p in self.parents)
        string += '], children=['
        string += ', '.join(str(c) for c in self.children)
        string += ']'

        return string

    def __eq__(self, other):
        if (not isinstance(other, Node)
           or (self.type != other.type)
           or (self.name != other.name)
           or (self.module != other.module)
           or (self.metadata != other.metadata)):
            return False
        return True

    def add_child(self, node):
        self.children.append(node)
        node.parents.append(self)
        return self


def traverse_depth_first(node, yield_root=True, level=0):
    """ Generator that traverses the dependency tree in a depth first manner and returns (level, node) tuples. |br|
    The general gist of this traversal is the following:

    .. code::

        1. Yield node N
        2. Foreach parent P of N (except previous node)
            3. Yield P
            4. Foreach parent PP of P
                5. Recurse to step 3 with P=PP
        6. Foreach child C of N
            7. Recurse to step 1 with C=N (and N=previous)

    Args:
        node (Node): Base node to start from
        yield_root(bool, optional): Whether to start by yielding the root node or not; Default **True**
        level (int, optional): Starting level value; Default **0**

    Returns:
        Generator : generator which returns `(leve, node)` tuples when iterated over

    Note:
        Strictly speaking, this is not depht first, as we yield the children whilst going down the path,
        but this is what we needed for lightnet.
    """
    def _traverse_depth_first(node, yield_root, prev, level):
        if yield_root:
            yield (level, node)
            yield from traverse_depth_first_parents(node, prev, level-1)

        for child in node.children:
            yield from _traverse_depth_first(child, True, [node], level+1)

    yield from _traverse_depth_first(node, yield_root, node.parents, level if yield_root else level-1)


def traverse_depth_first_parents(node, skip=None, level=0):
    """ Parent function of `traverse_depth_first` """
    if skip is None:
        skip = list()

    for p in node.parents:
        if p not in skip:
            yield (level, p)
            yield from traverse_depth_first_parents(p, list(), level-1)


def traverse_breadth_first(node, yield_root=True, level=0):
    """ Generator that traverses the dependency tree in a breadth first manner and returns (level, node) tuples. |br|
    The general gist of this traversal is the following:

    .. code::

        1. Foreach child C of node N
            2. Foreach parent P of C (except N)
                3. Recurse to step 2 with C=P and N=None
            4. Foreach parent P of C (except N)
                5. Yield P
            6. Yield C
        7. Foreach child C of node N
            8. Recurse to step 1 with N=C

    Args:
        node (Node): Base node to start from
        yield_root(bool, optional): Whether to start by yielding the root node or not; Default **True**
        level (int, optional): Starting level value; Default **0**

    Returns:
        Generator : generator which returns `(level, node)` tuples when iterated over
    """
    def _traverse_breadth_first(node, yield_root, level):
        if yield_root:
            yield (level, node)

        for child in node.children:
            yield from traverse_parents_breadth_first(child, [node], level)
            yield (level+1, child)

        for child in node.children:
            yield from traverse_breadth_first(child, False, level+1)

    yield from _traverse_breadth_first(node, yield_root, level if yield_root else level-1)


def traverse_parents_breadth_first(node, skip=None, level=0):
    """ Parent function of `traverse_breadth_first` """
    if skip is None:
        skip = list()

    for p in node.parents:
        if p not in skip:
            yield from traverse_parents_breadth_first(p, list(), level-1)

    for p in node.parents:
        if p not in skip:
            yield (level, p)



#
#   Lightnet optional dependencies
#   Copyright EAVISE
#
import logging

__all__ = ['pd', 'bb', 'cv2', 'Image', 'ImageOps']
log = logging.getLogger(__name__)

try:
    import pandas as pd
    import brambox as bb
except ModuleNotFoundError:
    log.warning('Brambox is not installed and thus all data functionality related to it cannot be used')
    pd = None
    bb = None

try:
    import cv2
except ModuleNotFoundError:
    log.warning('OpenCV is not installed and cannot be used')
    cv2 = None

try:
    from PIL import Image, ImageOps
except ModuleNotFoundError:
    log.warning('Pillow is not installed and cannot be used')
    Image, ImageOps = None, None



#
#   Node creation functions
#   Copyright EAVISE
#
import functools
from .._imports import *
from ._tree import *
from ._tree import traverse_parents_breadth_first
from ._type import *


__all__ = ['create_node', 'create_node_inverse']


# Recursive node creation
def create_node(element, graph, model, rootnode=None, parent=None):
    nodetype = get_node_type(element)
    create, out_dep, in_dep = node_func_map[nodetype]
    node = create(element, graph, model, rootnode, parent)
    deps = out_dep(element, graph, rootnode, node)

    # Check if rootnode
    if rootnode is None:
        rootnode = node

    # Loop through children and recurse
    for out in deps:
        if out in [o.name for o in graph.output]:
            raise StopIteration(out)

        for gn in graph.node:
            if out in gn.input:
                create_node(gn, graph, model, rootnode, node)

    return rootnode


def create_node_inverse(element, graph, model, rootnode, child):
    nodetype = get_node_type(element)
    create, out_dep, in_dep = node_func_map[nodetype]
    node = create(element, graph, model, rootnode, None)
    deps = in_dep(element, graph, rootnode, node)

    # StopNode
    if node is None:
        return

    # Check if node already exists
    if child is not None and node in child.parents:
        return [n for n in child.parents if node == n][0]
    if rootnode is not None:
        for _, c in traverse_depth_first(rootnode):
            if node == c:
                c.add_child(child)
                return c

    # Add child to node
    if child is not None:
        node.add_child(child)

    # Loop through parents and recurse
    for inp in deps:
        for gn in graph.node:
            if inp in gn.output:
                create_node_inverse(gn, graph, model, rootnode, node)

    return node


# Create specific nodes
def create_module_weight(nodetype, element, graph, model, rootnode, parent):
    name = [i for i in element.input if i.endswith('.weight')][0][:-7]
    try:
        module = model
        for p in name.split('.'):
            module = getattr(module, p)
    except AttributeError:
        log.debug(f'Could not get PyTorch module for [{name}]')
        module = None

    node = Node(nodetype, name, metadata=(set(element.input), set(element.output)), module=module)
    if parent is not None:
        parent.add_child(node)

    return node


def create_conv(element, graph, model, rootnode, parent):
    node = create_module_weight(NodeType.CONV, element, graph, model, rootnode, parent)

    if node.module.groups == 1:
        # Regular convolution
        return node
    elif node.module.groups == node.module.in_channels == node.module.out_channels:
        # Depthwise separable convolution
        return node
    else:
        # Grouped convolution
        raise NotImplementedError('ATen: Grouped Convolution (Only supporting DW-separable conv as dependency)')


def create_concat(element, graph, model, rootnode, parent):
    node = Node(NodeType.CONCAT, metadata=(set(element.input), set(element.output)))

    # Get concatenation axis (Constant input to concat operation)
    inputs = [i for i in element.input if len(i) > 0]
    for gn in graph.node:
        if gn.output[0] in inputs and gn.op_type == 'Constant':
            value = onnx_numpy_helper.to_array(onnx.helper.get_attribute_value(gn.attribute[0]))
            if value != 1:
                raise NotImplementedError(f'ATen: cat (Not concatenating on channel dimension [1])')

    deps = input_dep_all(element, graph, rootnode, node)
    for inp in deps:
        for gn in graph.node:
            if inp in gn.output:
                create_node_inverse(gn, graph, model, rootnode, node)

    if parent is not None and parent not in node.parents:
        parent.add_child(node)
    return node


def create_elemw_op(element, graph, model, rootnode, parent):
    attr = onnx.helper.printable_attribute(element.attribute[0])
    attr = attr.split('=')[-1].strip()[1:-1]

    if parent is not None:
        # Parent is None in create_node_inverse and in first iteration of create_node
        # elem_op should never be root of tree (is always conv), but this is probably brittle code
        raise NotImplementedError(f'ATen: {attr} (downward path)')

    node = Node(NodeType.ELEMW_OP, attr, metadata=(set(element.input), set(element.output)))

    # Search for a parent node from which we can get size information
    deps = input_dep_all(element, graph, rootnode, node)
    good_dep = None
    for inp in deps:
        for gn in graph.node:
            if inp in gn.output:
                try:
                    dep_node = create_node_inverse(gn, graph, model, None, None)
                except (NotImplementedError, StopIteration) as err:
                    dep_node = None

                if dep_node:
                    for _, n in traverse_parents_breadth_first(dep_node):
                        if n.type in {NodeType.CONV, NodeType.BATCHNORM, NodeType.CONCAT}:
                            good_dep = n
                            break

                    if good_dep is not None:
                        break
        if good_dep is not None:
            break

    # No suitable parent found
    if good_dep is None:
        raise NotImplementedError(f'ATen: {attr} (no readable upward dependency)')

    good_dep.children = list()
    good_dep.parents = list()
    good_dep.add_child(node)
    return node


def create_ignore(element, graph, model, rootnode, parent):
    name = onnx.helper.printable_attribute(element.attribute[0]).split('=')[-1].strip()
    node = Node(NodeType.IGNORE, name, metadata=(set(element.input), set(element.output)))

    if name == '"flip"':
        inputs = [i for i in element.input if len(i) > 0]
        for gn in graph.node:
            if gn.output[0] in inputs and gn.op_type == 'Constant':
                value = onnx_numpy_helper.to_array(onnx.helper.get_attribute_value(gn.attribute[0]))
                if value == 1:
                    raise NotImplementedError(f'ATen: flip (Flipping on channel dimension [1])')

    if parent is not None:
        parent.add_child(node)

    return node


def create_ignore_stop(element, graph, model, rootnode, parent):
    return None


# Return dependencies
def output_dep_all(element, graph, rootnode, node):
    return element.output


def input_dep_all(element, graph, rootnode, node):
    inputs = [i for i in element.input if len(i) > 0]

    # Remove graph inputs
    for i in graph.input:
        if i.name in inputs:
            inputs.remove(i.name)

    # Remove constants
    for gn in graph.node:
        if gn.output[0] in inputs and gn.op_type == 'Constant':
            inputs.remove(gn.output[0])

    return inputs


def dep_none(element, graph, rootnode, node):
    return list()


def output_dep_conv(element, graph, rootnode, node):
    if rootnode is None or node.module.groups != 1:
        return element.output
    else:
        return list()


# Map with 3 funcs (create, output_dep, input_dep) for each NodeType
# WARNING : create/input_dep cannot count on rootnode being correct, because it is not when constructing tree inversely (cat/elemw)
node_func_map = {
    NodeType.CONV: (
        create_conv,
        output_dep_conv,
        dep_none
    ),
    NodeType.BATCHNORM: (
        functools.partial(create_module_weight, NodeType.BATCHNORM),
        output_dep_all,
        dep_none
    ),
    NodeType.CONCAT: (
        create_concat,
        output_dep_all,
        dep_none,        # Gets input deps in create function
    ),
    NodeType.ELEMW_OP: (
        create_elemw_op,
        dep_none,
        dep_none,        # Gets input deps in create function
    ),
    NodeType.IGNORE: (
        create_ignore,
        output_dep_all,
        input_dep_all
    ),
    NodeType.IGNORE_STOP: (
        create_ignore_stop,
        dep_none,
        dep_none
    ),
}



#
#   Test if network forward function runs
#   Copyright EAVISE
#

import inspect
import pytest
import torch
import lightnet as ln

classification_networks = ['Darknet', 'Darknet19', 'Darknet53', 'MobileDarknet19', 'MobilenetV1', 'MobilenetV2']
anchor_detection_networks = ['DYolo', 'MobilenetYolo', 'MobileYoloV2', 'MobileYoloV2Upsample', 'TinyYoloV2', 'TinyYoloV3', 'YoloV2', 'YoloV2Upsample', 'YoloV3', 'Yolt']
corner_detection_networks = ['Cornernet', 'CornernetSqueeze']
special_networks = ['YoloFusion']


@pytest.fixture(scope='module')
def input_tensor():
    def _input_tensor(dimension, channels=3, batch=1):
        return torch.rand(batch, channels, dimension, dimension)
    return _input_tensor


# Base classification networks
@pytest.mark.parametrize('network', classification_networks)
def test_classification_cpu(network, input_tensor):
    uut = getattr(ln.models, network)(1000).eval()
    it = input_tensor(uut.inner_stride)

    output_tensor = uut(it)
    assert output_tensor.dim() == 2
    assert output_tensor.shape[0] == it.shape[0]
    assert output_tensor.shape[1] == uut.num_classes


@pytest.mark.parametrize('network', classification_networks)
@pytest.mark.cuda
@pytest.mark.skipif(not torch.cuda.is_available(), reason='CUDA not available')
def test_classification_cuda(network, input_tensor):
    uut = getattr(ln.models, network)(1000).to('cuda')
    it = input_tensor(uut.inner_stride, batch=2).to('cuda')

    output_tensor = uut(it)
    assert output_tensor.dim() == 2
    assert output_tensor.shape[0] == it.shape[0]
    assert output_tensor.shape[1] == uut.num_classes


# Anchor detection networks
@pytest.mark.parametrize('network', anchor_detection_networks)
def test_anchor_detection_cpu(network, input_tensor):
    uut = getattr(ln.models, network)(20).eval()
    it = input_tensor(uut.inner_stride)

    output_tensor = uut(it)
    if isinstance(output_tensor, torch.Tensor):
        assert output_tensor.dim() == 4
        assert output_tensor.shape[0] == it.shape[0]
        assert output_tensor.shape[1] == len(uut.anchors) * (5 + uut.num_classes)
        assert output_tensor.shape[2] == it.shape[2] // uut.stride
        assert output_tensor.shape[3] == it.shape[3] // uut.stride
    else:
        for i, tensor in enumerate(output_tensor):
            assert tensor.dim() == 4
            assert tensor.shape[0] == it.shape[0]
            assert tensor.shape[1] == len(uut.anchors[i]) * (5 + uut.num_classes)
            assert tensor.shape[2] == it.shape[2] // uut.stride[i]
            assert tensor.shape[3] == it.shape[3] // uut.stride[i]


@pytest.mark.parametrize('network', anchor_detection_networks)
@pytest.mark.cuda
@pytest.mark.skipif(not torch.cuda.is_available(), reason='CUDA not available')
def test_anchor_detection_cuda(network, input_tensor):
    uut = getattr(ln.models, network)(20).to('cuda')
    it = input_tensor(uut.inner_stride, batch=2).to('cuda')

    output_tensor = uut(it)
    if isinstance(output_tensor, torch.Tensor):
        assert output_tensor.dim() == 4
        assert output_tensor.shape[0] == it.shape[0]
        assert output_tensor.shape[1] == len(uut.anchors) * (5 + uut.num_classes)
        assert output_tensor.shape[2] == it.shape[2] // uut.stride
        assert output_tensor.shape[3] == it.shape[3] // uut.stride
    else:
        for i, tensor in enumerate(output_tensor):
            assert tensor.dim() == 4
            assert tensor.shape[0] == it.shape[0]
            assert tensor.shape[1] == len(uut.anchors[i]) * (5 + uut.num_classes)
            assert tensor.shape[2] == it.shape[2] // uut.stride[i]
            assert tensor.shape[3] == it.shape[3] // uut.stride[i]


# Corner detection networks
@pytest.mark.parametrize('network', corner_detection_networks)
def test_corner_detection_cpu(network, input_tensor):
    uut = getattr(ln.models, network)(20).eval()
    it = input_tensor(uut.inner_stride)

    output_tensor = uut(it)
    if isinstance(output_tensor, torch.Tensor):
        assert output_tensor.dim() == 4
        assert output_tensor.shape[0] == it.shape[0]
        assert output_tensor.shape[1] == (uut.num_classes + 3) * 2
        assert output_tensor.shape[2] == it.shape[2] // uut.stride
        assert output_tensor.shape[3] == it.shape[3] // uut.stride
    else:
        for tensor in output_tensor:
            assert tensor.dim() == 4
            assert tensor.shape[0] == it.shape[0]
            assert tensor.shape[1] == (uut.num_classes + 3) * 2
            assert tensor.shape[2] == it.shape[2] // uut.stride
            assert tensor.shape[3] == it.shape[3] // uut.stride


@pytest.mark.parametrize('network', corner_detection_networks)
@pytest.mark.cuda
@pytest.mark.skipif(not torch.cuda.is_available(), reason='CUDA not available')
def test_corner_detection_cuda(network, input_tensor):
    uut = getattr(ln.models, network)(20).to('cuda')
    it = input_tensor(uut.inner_stride, batch=2).to('cuda')

    output_tensor = uut(it)
    if isinstance(output_tensor, torch.Tensor):
        assert output_tensor.dim() == 4
        assert output_tensor.shape[0] == it.shape[0]
        assert output_tensor.shape[1] == (uut.num_classes + 3) * 2
        assert output_tensor.shape[2] == it.shape[2] // uut.stride
        assert output_tensor.shape[3] == it.shape[3] // uut.stride
    else:
        for tensor in output_tensor:
            assert tensor.dim() == 4
            assert tensor.shape[0] == it.shape[0]
            assert tensor.shape[1] == (uut.num_classes + 3) * 2
            assert tensor.shape[2] == it.shape[2] // uut.stride
            assert tensor.shape[3] == it.shape[3] // uut.stride


# YoloFusion
def test_yolofusion_cpu(input_tensor):
    it = input_tensor(ln.models.YoloFusion.inner_stride, 4)

    for fusion in (0, 1, 10, 22, 27):
        uut = ln.models.YoloFusion(20, fuse_layer=fusion).eval()
        output_tensor = uut(it)
        assert output_tensor.dim() == 4
        assert output_tensor.shape[0] == it.shape[0]
        assert output_tensor.shape[1] == len(uut.anchors) * (5 + uut.num_classes)
        assert output_tensor.shape[2] == it.shape[2] // uut.stride
        assert output_tensor.shape[3] == it.shape[3] // uut.stride


@pytest.mark.cuda
@pytest.mark.skipif(not torch.cuda.is_available(), reason='CUDA not available')
def test_yolofusion_cuda(input_tensor):
    it = input_tensor(ln.models.YoloFusion.inner_stride * 2, 4).to('cuda')

    for fusion in (0, 1, 10, 22, 27):
        uut = ln.models.YoloFusion(20, fuse_layer=fusion).to('cuda')
        output_tensor = uut(it)
        assert output_tensor.dim() == 4
        assert output_tensor.shape[0] == it.shape[0]
        assert output_tensor.shape[1] == len(uut.anchors) * (5 + uut.num_classes)
        assert output_tensor.shape[2] == it.shape[2] // uut.stride
        assert output_tensor.shape[3] == it.shape[3] // uut.stride


# All networks tested?
def test_all_networks_tested():
    networks = [
        net for net in dir(ln.models)
        if (inspect.isclass(getattr(ln.models, net)))
        and (issubclass(getattr(ln.models, net), torch.nn.Module))
    ]

    tested_networks = set(
        anchor_detection_networks
        + corner_detection_networks
        + classification_networks
        + special_networks
    )
    for net in networks:
        if net not in tested_networks:
            raise NotImplementedError(f'Network [{net}] is not being tested!')



#
#   Lightnet postprocessing reverse fit
#   Copyright EAVISE
#

from ..util import BaseTransform

__all__ = ['ReverseCrop', 'ReverseLetterbox', 'ReversePad']


class ReverseCrop(BaseTransform):
    """ Performs a reverse :class:`~lightnet.data.transform.Crop` operation on the bounding boxes, so that the bounding box coordinates are relative to the original image dimensions.

    Args:
        network_size (tuple): Tuple containing the width and height of the images going in the network
        image_size (tuple, callable or dict-like): Width and height of the original images (See Note)

    Returns:
        pandas.DataFrame: brambox dataframe.

    Note:
        The `image_size` argument can be one of three different types:

        - tuple <width, height> : The same image size will be used for the entire dataframe
        - callable : The argument will be called with the image column name and must return a (width, height) tuple
        - dict-like : This is similar to the callable, but instead of calling the argument, it will use dictionary accessing (self.image_size[img_name])

        Note that if your dimensions are the same for all images, it is faster to pass a tuple,
        as the transformation will be applied to the entire dataframe at once as opposed to grouping it per image and applying the tranform to each group individually.

    Warning:
        This post-processing only works when center-cropping images.
        Make sure to set the `center` argument to **True** in your :class:`~lightnet.data.transform.Crop` pre-processing.
    """
    def __init__(self, network_size, image_size):
        super().__init__()
        self.network_size = network_size
        self.image_size = image_size

    def forward(self, boxes):
        if isinstance(self.image_size, (list, tuple)):
            net_w, net_h = self.network_size
            im_w, im_h = self.image_size
            scale, pad = self._get_params(net_w, net_h, im_w, im_h)
            return self._transform(boxes.copy(), scale, pad)

        return boxes.groupby('image').apply(self._apply_transform)

    def _apply_transform(self, boxes):
        net_w, net_h = self.network_size
        if callable(self.image_size):
            im_w, im_h = self.image_size(boxes.name)
        else:
            im_w, im_h = self.image_size[boxes.name]

        scale, pad = self._get_params(net_w, net_h, im_w, im_h)
        return self._transform(boxes.copy(), scale, pad)

    @staticmethod
    def _get_params(net_w, net_h, im_w, im_h):
        if net_w / im_w >= net_h / im_h:
            scale = im_w / net_w
            dx = 0
            dy = int(im_h / scale - net_h + 0.5) // 2
        else:
            scale = im_h / net_h
            dx = int(im_w / scale - net_w + 0.5) // 2
            dy = 0

        return scale, (dx, dy)

    @staticmethod
    def _transform(boxes, scale, pad):
        boxes.x_top_left += pad[0]
        boxes.y_top_left += pad[1]

        boxes.x_top_left *= scale
        boxes.y_top_left *= scale
        boxes.width *= scale
        boxes.height *= scale

        return boxes


class ReverseLetterbox(BaseTransform):
    """ Performs a reverse :class:`~lightnet.data.transform.Letterbox` operation on the bounding boxes, so that the bounding box coordinates are relative to the original image dimensions.

    Args:
        network_size (tuple): Tuple containing the width and height of the images going in the network
        image_size (tuple, callable or dict-like): Width and height of the original images (See Note)

    Returns:
        pandas.DataFrame: brambox dataframe.

    Note:
        The `image_size` argument can be one of three different types:

        - tuple <width, height> : The same image size will be used for the entire dataframe
        - callable : The argument will be called with the image column name and must return a (width, height) tuple
        - dict-like : This is similar to the callable, but instead of calling the argument, it will use dictionary accessing (self.image_size[img_name])

        Note that if your dimensions are the same for all images, it is faster to pass a tuple,
        as the transformation will be applied to the entire dataframe at once as opposed to grouping it per image and applying the tranform to each group individually.
    """
    def __init__(self, network_size, image_size):
        super().__init__()
        self.network_size = network_size
        self.image_size = image_size

    def forward(self, boxes):
        if isinstance(self.image_size, (list, tuple)):
            net_w, net_h = self.network_size
            im_w, im_h = self.image_size
            scale, pad = self._get_params(net_w, net_h, im_w, im_h)
            return self._transform(boxes.copy(), scale, pad)

        return boxes.groupby('image').apply(self._apply_transform)

    def _apply_transform(self, boxes):
        net_w, net_h = self.network_size
        if callable(self.image_size):
            im_w, im_h = self.image_size(boxes.name)
        else:
            im_w, im_h = self.image_size[boxes.name]

        scale, pad = self._get_params(net_w, net_h, im_w, im_h)
        return self._transform(boxes.copy(), scale, pad)

    @staticmethod
    def _get_params(net_w, net_h, im_w, im_h):
        if im_w == net_w and im_h == net_h:
            scale = 1
        elif im_w / net_w >= im_h / net_h:
            scale = im_w/net_w
        else:
            scale = im_h/net_h
        pad = (net_w - im_w/scale) // 2, (net_h - im_h/scale) // 2

        return scale, pad

    @staticmethod
    def _transform(boxes, scale, pad):
        boxes.x_top_left -= pad[0]
        boxes.y_top_left -= pad[1]

        boxes.x_top_left *= scale
        boxes.y_top_left *= scale
        boxes.width *= scale
        boxes.height *= scale

        return boxes


class ReversePad(BaseTransform):
    """ Performs a reverse :class:`~lightnet.data.transform.Pad` operation on the bounding boxes, so that the bounding box coordinates are relative to the original image dimensions.

    Args:
        network_factor (int or tuple): Tuple containing the factor the width and height need to match
        image_size (tuple, callable or dict-like): Width and height of the original images (See Note)

    Returns:
        pandas.DataFrame: brambox dataframe.

    Note:
        The `image_size` argument can be one of three different types:

        - tuple <width, height> : The same image size will be used for the entire dataframe
        - callable : The argument will be called with the image column name and must return a (width, height) tuple
        - dict-like : This is similar to the callable, but instead of calling the argument, it will use dictionary accessing (self.image_size[img_name])

        Note that if your dimensions are the same for all images, it is faster to pass a tuple,
        as the transformation will be applied to the entire dataframe at once as opposed to grouping it per image and applying the tranform to each group individually.
    """
    def __init__(self, network_factor, image_size):
        super().__init__()
        self.network_factor = network_factor
        self.image_size = image_size

    def forward(self, boxes):
        if isinstance(self.image_size, (list, tuple)):
            im_w, im_h = self.image_size
            self._get_params(im_w, im_h)
            return self._transform(boxes.copy(), self.pad)

        return boxes.groupby('image').apply(self._apply_transform)

    def _get_params(self, im_w, im_h):
        if isinstance(self.network_factor, int):
            net_w, net_h = self.network_factor, self.network_factor
        else:
            net_w, net_h = self.network_factor

        self.pad = ((net_w - (im_w % net_w)) % net_w) // 2, ((net_h - (im_h % net_h)) % net_h) // 2

    def _apply_transform(self, boxes):
        if callable(self.image_size):
            im_w, im_h = self.image_size(boxes.name)
        else:
            im_w, im_h = self.image_size[boxes.name]
        self._get_params(im_w, im_h)

        return self._transform(boxes.copy(), self.pad)

    @staticmethod
    def _transform(boxes, pad):
        boxes.x_top_left -= pad[0]
        boxes.y_top_left -= pad[1]
        return boxes



#
#   Image and annotations preprocessing for lightnet networks
#   The image transformations work with both Pillow and OpenCV images or with PyTorch tensors
#   The annotation transformations work with brambox 2 annotations dataframes
#   Copyright EAVISE
#

from ._fit import *
from ._augment import *
from ._util import *



#
#   Lightnet dataset that works with brambox annotations
#   Copyright EAVISE
#

import os
import copy
import logging
from PIL import Image
import numpy as np
import lightnet.data as lnd

try:
    import brambox as bb
except ImportError:
    bb = None

__all__ = ['BramboxDataset']
log = logging.getLogger(__name__)


class BramboxDataset(lnd.Dataset):
    """ Dataset for any brambox annotations.

    Args:
        annotations (dataframe): Dataframe containing brambox annotations
        input_dimension (tuple, optional): (width,height) tuple with default dimensions of the network; Default **None**
        class_label_map (list): List of class_labels
        identify (function, optional): Lambda/function to get image based of annotation filename or image id; Default **replace/add .png extension to filename/id**
        transform (torchvision.transforms.Compose): Transformation pipeline
        anno_transform (torchvision.transforms.Compose): Annotation transformation pipeline

    Note:
        If you only pass a ``transform`` pipeline, it will be called with both your image and annotations as a tuple.
        However, if you pass both a ``transform`` and ``anno_transform`` pipeline, it will first run your image through the ``transform`` pipeline,
        and then your annotations through ``anno_transform`` (legacy behaviour).

    Note:
        This dataset opens images with the Pillow library
    """
    def __init__(self, annotations, input_dimension=None, class_label_map=None, identify=None, transform=None, anno_transform=None):
        if bb is None:
            raise ImportError('Brambox needs to be installed to use this dataset')
        super().__init__(input_dimension)

        self.annos = annotations
        self.keys = self.annos.image.cat.categories
        self.transform = transform
        self.anno_transform = anno_transform

        if callable(identify):
            self.id = identify
        else:
            self.id = lambda name: os.path.splitext(name)[0] + '.png'

        # Add class_ids
        if class_label_map is None:
            log.warning(f'No class_label_map given, generating it by sorting unique class labels from data alphabetically, which is not always deterministic behaviour')
            class_label_map = list(np.sort(self.annos.class_label.unique()))
        self.annos['class_id'] = self.annos.class_label.map(dict((l, i) for i, l in enumerate(class_label_map)))

    def __len__(self):
        return len(self.keys)

    @lnd.Dataset.resize_getitem
    def __getitem__(self, index):
        """ Get transformed image and annotations based of the index of ``self.keys``

        Args:
            index (int): index of the ``self.keys`` list containing all the image identifiers of the dataset.

        Returns:
            tuple: (transformed image, list of transformed brambox boxes)
        """
        if index >= len(self):
            raise IndexError(f'list index out of range [{index}/{len(self)-1}]')

        # Load
        img = Image.open(self.id(self.keys[index]))
        anno = bb.util.select_images(self.annos, [self.keys[index]])

        # Transform
        if self.transform is not None and self.anno_transform is None:
            (img, anno) = self.transform((img, anno))
        elif self.transform is not None:
            img = self.transform(img)
            if self.anno_transform is not None:
                anno = self.anno_transform(anno)

        return img, anno
